"""Test the router for the Job Script Template resource."""

import json

import httpx
import pytest
from fastapi import status

from jobbergate_api.apps.job_script_templates.constants import WORKFLOW_FILE_NAME
from jobbergate_api.apps.job_script_templates.schemas import (
    JobTemplateBaseDetailedView,
    JobTemplateListView,
)
from jobbergate_api.apps.permissions import Permissions
from jobbergate_api.apps.services import ServiceError

# Not using the synth_session fixture in a route that needs the database is unsafe
pytest.mark.usefixtures("synth_session")


@pytest.mark.parametrize("permission", (Permissions.ADMIN, Permissions.JOB_TEMPLATES_CREATE))
async def test_create_job_template__success(
    permission,
    client,
    fill_job_template_data,
    inject_security_header,
    synth_services,
):
    """Test a job template can be created."""
    payload = fill_job_template_data(
        identifier="create-template",
        name="Test Template",
        description="This is a test template",
        template_vars={"foo": "bar"},
    )

    tester_email = payload.pop("owner_email")
    inject_security_header(tester_email, permission)

    response = await client.post("jobbergate/job-script-templates", json=payload)
    assert response.status_code == 201, f"Create failed: {response.text}"
    response_data = response.json()

    assert response_data["name"] == payload["name"]
    assert response_data["description"] == payload["description"]
    assert response_data["identifier"] == payload["identifier"]
    assert response_data["template_vars"] == payload["template_vars"]
    assert response_data["owner_email"] == tester_email
    assert "template_files" not in response_data
    assert "workflow_files" not in response_data

    # Make sure the data was actually inserted into the database
    assert (await synth_services.crud.template.count()) == 1
    instance = await synth_services.crud.template.get(response_data["id"])
    assert instance is not None
    assert instance.identifier == "create-template"
    assert instance.name == "Test Template"
    assert instance.description == "This is a test template"
    assert instance.template_vars == {"foo": "bar"}

    # Make sure that the data can be retrieved with a GET request
    inject_security_header(tester_email, Permissions.JOB_TEMPLATES_READ)
    response = await client.get(f"jobbergate/job-script-templates/{instance.id}")
    assert response.status_code == 200
    response_data = response.json()
    assert response_data["identifier"] == "create-template"
    assert response_data["name"] == "Test Template"
    assert response_data["description"] == "This is a test template"
    assert response_data["template_vars"] == {"foo": "bar"}


async def test_create_job_template__fails_if_name_is_empty(
    client,
    fill_job_template_data,
    inject_security_header,
    synth_services,
):
    payload = fill_job_template_data(
        identifier="create-template",
        name="",
        description="This is a test template",
        template_vars={"foo": "bar"},
    )

    tester_email = payload.pop("owner_email")
    inject_security_header(tester_email, Permissions.JOB_TEMPLATES_CREATE)

    response = await client.post("jobbergate/job-script-templates", json=payload)
    assert response.status_code == 422
    assert "Cannot be an empty string" in response.text
    assert (await synth_services.crud.template.count()) == 0


async def test_create_job_template__coerces_empty_identifier_to_None(
    client,
    fill_job_template_data,
    inject_security_header,
    synth_services,
):
    payload = fill_job_template_data(
        identifier="",
        name="Test Template",
        description="This is a test template",
        template_vars={"foo": "bar"},
    )

    tester_email = payload.pop("owner_email")
    inject_security_header(tester_email, Permissions.JOB_TEMPLATES_CREATE)

    response = await client.post("jobbergate/job-script-templates", json=payload)
    assert response.status_code == 201, f"Create failed: {response.text}"
    response_data = response.json()

    assert response_data["identifier"] is None


async def test_create_job_template__fail_unauthorized(client, fill_job_template_data, synth_services):
    """Test that the job template creation fails if the user is unauthorized."""
    payload = fill_job_template_data()
    response = await client.post("jobbergate/job-script-templates", json=payload)
    assert response.status_code == 401
    assert (await synth_services.crud.template.count()) == 0


async def test_create_job_template__fail_identifier_already_exists(
    client,
    fill_job_template_data,
    inject_security_header,
    synth_services,
):
    """Test that the job template creation fails if the identifier already exists."""
    payload = fill_job_template_data(identifier="duplicated-template")

    tester_email = payload.pop("owner_email")
    inject_security_header(tester_email, Permissions.JOB_TEMPLATES_CREATE)

    response = await client.post("jobbergate/job-script-templates", json=payload)
    assert response.status_code == 201
    assert (await synth_services.crud.template.count()) == 1

    response = await client.post("jobbergate/job-script-templates", json=payload)
    assert response.status_code == 409
    assert (await synth_services.crud.template.count()) == 1


async def test_create_job_template__fail_missing_name(
    client,
    fill_job_template_data,
    inject_security_header,
    synth_services,
):
    """Test that the job template creation fails if a required field is missing."""
    payload = fill_job_template_data()

    payload.pop("name")

    tester_email = payload.pop("owner_email")
    inject_security_header(tester_email, Permissions.JOB_TEMPLATES_CREATE)

    response = await client.post("jobbergate/job-script-templates", json=payload)

    assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY

    assert (await synth_services.crud.template.count()) == 0


@pytest.mark.parametrize(
    "is_owner, permissions",
    [
        (True, [Permissions.JOB_TEMPLATES_UPDATE]),
        (False, [Permissions.ADMIN]),
        (False, [Permissions.JOB_TEMPLATES_UPDATE, Permissions.MAINTAINER]),
    ],
)
async def test_update_job_template__success(
    client,
    fill_job_template_data,
    inject_security_header,
    tester_email,
    synth_services,
    is_owner,
    permissions,
):
    instance = await synth_services.crud.template.create(**fill_job_template_data())

    requester_email = tester_email if is_owner else "another_" + tester_email

    payload = {
        "name": "new-name",
        "identifier": "new-identifier",
        "description": "new-description",
        "template_vars": {"new": "value"},
    }

    inject_security_header(requester_email, *permissions)
    response = await client.put(
        f"jobbergate/job-script-templates/{instance.id}",
        json=payload,
    )

    assert response.status_code == 200, f"Update failed: {response.text}"
    response_data = response.json()
    assert response_data["name"] == payload["name"]
    assert response_data["description"] == payload["description"]
    assert response_data["identifier"] == payload["identifier"]
    assert response_data["template_vars"] == payload["template_vars"]


async def test_update_job_template__fail_not_found(
    client,
    tester_email,
    inject_security_header,
    synth_session,
):
    job_template_id = 0
    payload = {
        "name": "new-name",
        "identifier": "new-identifier",
        "description": "new-description",
        "template_vars": {"new": "value"},
    }
    inject_security_header(tester_email, Permissions.JOB_TEMPLATES_UPDATE)
    response = await client.put(f"jobbergate/job-script-templates/{job_template_id}", json=payload)
    assert response.status_code == 404


async def test_update_job_template__fail_unauthorized(client):
    job_template_id = 0
    payload = {
        "name": "new-name",
        "identifier": "new-identifier",
        "description": "new-description",
        "template_vars": {"new": "value"},
    }
    response = await client.put(f"jobbergate/job-script-templates/{job_template_id}", json=payload)
    assert response.status_code == 401


async def test_update_job_template__forbidden(
    client,
    fill_job_template_data,
    inject_security_header,
    tester_email,
    synth_services,
):
    instance = await synth_services.crud.template.create(**fill_job_template_data())

    owner_email = tester_email
    requester_email = "another_" + owner_email

    payload = {
        "name": "new-name",
        "identifier": "new-identifier",
        "description": "new-description",
        "template_vars": {"new": "value"},
    }

    inject_security_header(requester_email, Permissions.JOB_TEMPLATES_UPDATE)
    response = await client.put(
        f"jobbergate/job-script-templates/{instance.id}",
        json=payload,
    )

    assert response.status_code == 403


@pytest.mark.parametrize("identification_field", ("id", "identifier"))
@pytest.mark.parametrize("permission", (Permissions.ADMIN, Permissions.JOB_TEMPLATES_READ))
async def test_get_job_template__success(
    permission,
    identification_field,
    client,
    inject_security_header,
    fill_job_template_data,
    synth_services,
):
    payload = fill_job_template_data(
        identifier=f"get-template-{identification_field}",
        template_vars={"foo": "bar"},
    )
    instance = await synth_services.crud.template.create(**payload)

    identification = getattr(instance, identification_field)

    inject_security_header(instance.owner_email, permission)
    response = await client.get(f"jobbergate/job-script-templates/{identification}")
    assert response.status_code == 200, f"Get failed: {response.text}"
    response_data = response.json()

    assert response_data["name"] == payload["name"]
    assert response_data["description"] == payload["description"]
    assert response_data["identifier"] == payload["identifier"]
    assert response_data["template_vars"] == payload["template_vars"]
    assert response_data["owner_email"] == payload["owner_email"]


@pytest.mark.parametrize("identification_field", ("id", "identifier"))
@pytest.mark.parametrize(
    "is_owner, permissions",
    [
        (True, [Permissions.JOB_TEMPLATES_DELETE]),
        (False, [Permissions.ADMIN]),
        (False, [Permissions.JOB_TEMPLATES_DELETE, Permissions.MAINTAINER]),
    ],
)
async def test_delete_job_template__success(
    identification_field,
    client,
    tester_email,
    inject_security_header,
    fill_job_template_data,
    synth_services,
    is_owner,
    permissions,
):
    payload = fill_job_template_data(
        identifier=f"delete-template-{identification_field}",
        template_vars={"foo": "bar"},
    )
    instance = await synth_services.crud.template.create(**payload)

    requester_email = tester_email if is_owner else "another_" + tester_email

    inject_security_header(requester_email, *permissions)
    identification = getattr(instance, identification_field)
    response = await client.delete(f"jobbergate/job-script-templates/{identification}")
    assert response.status_code == 204, f"Delete failed: {response.text}"

    assert (await synth_services.crud.template.count()) == 0


async def test_delete_job_template__fail_not_found(
    client,
    tester_email,
    inject_security_header,
    synth_session,
):
    job_template_id = 0
    inject_security_header(tester_email, Permissions.JOB_TEMPLATES_DELETE)
    response = await client.delete(f"jobbergate/job-script-templates/{job_template_id}")
    assert response.status_code == 404


@pytest.mark.parametrize("identification_field", ("id", "identifier"))
async def test_delete_job_template__forbidden(
    identification_field,
    client,
    tester_email,
    inject_security_header,
    fill_job_template_data,
    synth_services,
):
    owner_email = tester_email
    requester_email = "another_" + owner_email

    payload = fill_job_template_data(
        identifier=f"delete-template-forbidden-{identification_field}",
        template_vars={"foo": "bar"},
    )
    instance = await synth_services.crud.template.create(**payload)

    inject_security_header(requester_email, Permissions.JOB_TEMPLATES_DELETE)
    identification = getattr(instance, identification_field)
    response = await client.delete(f"jobbergate/job-script-templates/{identification}")
    assert response.status_code == 403

    assert (await synth_services.crud.template.count()) == 1


@pytest.mark.parametrize("permission", (Permissions.ADMIN, Permissions.JOB_TEMPLATES_CREATE))
async def test_clone_job_template__success(
    permission,
    client,
    fill_job_template_data,
    inject_security_header,
    tester_email,
    synth_services,
):
    original_instance = await synth_services.crud.template.create(
        **fill_job_template_data(owner_email=tester_email, identifier="test-identifier")
    )
    parent_id = original_instance.id
    await synth_services.file.template.upsert(
        parent_id=parent_id,
        filename="test_template.py.j2",
        upload_content="print('dummy file data')",
        file_type="ENTRYPOINT",
    )
    await synth_services.file.template.upsert(
        parent_id=parent_id,
        filename="test_template.j2",
        upload_content="support dummy file data",
        file_type="SUPPORT",
    )
    await synth_services.file.workflow.upsert(
        parent_id=parent_id,
        filename="jobbergate.py",
        upload_content="print('dummy file data')",
    )

    new_owner_email = "new_" + tester_email

    inject_security_header(new_owner_email, permission)
    response = await client.post(f"jobbergate/job-script-templates/clone/{original_instance.id}")

    assert response.status_code == 201, f"Clone failed: {response.text}"
    response_data = response.json()

    cloned_id = response_data["id"]

    assert cloned_id != original_instance.id
    assert response_data["name"] == original_instance.name
    assert response_data["description"] == original_instance.description
    assert response_data["identifier"] is None
    assert response_data["template_vars"] == original_instance.template_vars
    assert response_data["owner_email"] == new_owner_email
    assert response_data["cloned_from_id"] == original_instance.id

    assert {f["filename"] for f in response_data["template_files"]} == {
        "test_template.py.j2",
        "test_template.j2",
    }
    assert {f["filename"] for f in response_data["workflow_files"]} == {"jobbergate.py"}


async def test_clone_job_template__replace_base_values(
    client,
    fill_job_template_data,
    inject_security_header,
    tester_email,
    synth_services,
):
    original_instance = await synth_services.crud.template.create(
        **fill_job_template_data(owner_email=tester_email, identifier="test-identifier")
    )

    new_owner_email = "new_" + tester_email

    payload = {
        "name": "new_name",
        "description": "new_description",
        "identifier": "new_identifier",
        "template_vars": {"new": "value"},
    }

    inject_security_header(new_owner_email, Permissions.JOB_TEMPLATES_CREATE)
    response = await client.post(
        f"jobbergate/job-script-templates/clone/{original_instance.id}", json=payload
    )

    assert response.status_code == 201, f"Clone failed: {response.text}"
    response_data = response.json()
    assert response_data["name"] == payload["name"]
    assert response_data["description"] == payload["description"]
    assert response_data["identifier"] == payload["identifier"]
    assert response_data["template_vars"] == payload["template_vars"]
    assert response_data["owner_email"] == new_owner_email
    assert response_data["cloned_from_id"] == original_instance.id


async def test_clone_job_template__fail_unauthorized(
    client,
    fill_job_template_data,
    synth_services,
):
    original_instance = await synth_services.crud.template.create(**fill_job_template_data())

    response = await client.post(f"jobbergate/job-script-templates/clone/{original_instance.id}")

    assert response.status_code == 401


async def test_clone_job_template__fail_not_found(
    client,
    inject_security_header,
    tester_email,
    synth_services,
):
    assert (await synth_services.crud.template.count()) == 0
    inject_security_header(tester_email, Permissions.JOB_TEMPLATES_CREATE)
    response = await client.post("jobbergate/job-script-templates/clone/0")

    assert response.status_code == 404


async def test_clone_job_template__fail_conflict(
    client,
    fill_job_template_data,
    inject_security_header,
    tester_email,
    synth_services,
):
    identifier = "test-identifier"
    original_instance = await synth_services.crud.template.create(
        **fill_job_template_data(owner_email=tester_email, identifier=identifier)
    )

    inject_security_header(tester_email, Permissions.JOB_TEMPLATES_CREATE)
    response = await client.post(
        f"jobbergate/job-script-templates/clone/{original_instance.id}", json={"identifier": identifier}
    )

    assert response.status_code == 409


class TestListJobTemplates:
    """Test the list endpoint."""

    @pytest.fixture(scope="function")
    async def job_templates_list(self, synth_services, fill_all_job_template_data):
        data = fill_all_job_template_data(
            {
                "name": "name-1",
                "description": "desc-1",
                "identifier": "identifier-1",
                "template_vars": {"foo-1": "bar-1"},
            },
            {
                "name": "name-2",
                "description": "desc-2",
                "identifier": "identifier-2",
                "template_vars": {"foo-1": "bar-2"},
            },
            {
                "name": "name-3",
                "description": "desc-3",
                "template_vars": {"foo-1": "bar-3"},
            },
            {
                "name": "name-4",
                "description": "desc-4",
                "template_vars": {"foo-1": "bar-4"},
                "owner_email": "test-test@pytest.com",
            },
            {
                "name": "name-5",
                "description": "desc-5",
                "template_vars": {"foo-1": "bar-5"},
                "is_archived": True,
            },
        )
        for item in data:
            await synth_services.crud.template.create(**item)
        yield data

    @pytest.mark.parametrize("permission", (Permissions.ADMIN, Permissions.JOB_TEMPLATES_READ))
    async def test_list_job_templates__all_success(
        self,
        permission,
        client,
        tester_email,
        inject_security_header,
        job_templates_list,
    ):
        inject_security_header(tester_email, permission)
        response = await client.get(
            "jobbergate/job-script-templates",
            params={"include_null_identifier": True, "include_archived": True, "sort_field": "id"},
        )
        assert response.status_code == 200, f"List failed: {response.text}"

        response_data = response.json()
        assert response_data["total"] == len(job_templates_list)
        assert response_data["page"] == 1
        assert response_data["size"] == 50
        assert response_data["pages"] == 1

        for response_item, expected_item in zip(response_data["items"], job_templates_list):
            assert response_item.get("identifier") == expected_item.get("identifier")
            assert response_item["name"] == expected_item["name"]
            assert response_item["description"] == expected_item["description"]
            assert response_item["owner_email"] == expected_item["owner_email"]
            assert response_item["is_archived"] == expected_item["is_archived"]

    async def test_list_job_templates__ignore_archived(
        self,
        client,
        tester_email,
        inject_security_header,
        job_templates_list,
    ):
        inject_security_header(tester_email, Permissions.JOB_TEMPLATES_READ)
        response = await client.get("jobbergate/job-script-templates?include_null_identifier=True")
        assert response.status_code == 200, f"List failed: {response.text}"

        response_data = response.json()

        expected_names = {i["name"] for i in job_templates_list if i["is_archived"] is False}
        response_names = {i["name"] for i in response_data["items"]}

        assert response_data["total"] == len(expected_names)
        assert expected_names == response_names

    async def test_list_job_templates__user_only(
        self,
        client,
        tester_email,
        inject_security_header,
        job_templates_list,
    ):
        inject_security_header(tester_email, Permissions.JOB_TEMPLATES_READ)
        response = await client.get(
            "jobbergate/job-script-templates?user_only=True&include_null_identifier=True&include_archived=True"
        )
        assert response.status_code == 200, f"List failed: {response.text}"

        response_data = response.json()

        expected_names = {i["name"] for i in job_templates_list if i["owner_email"] == tester_email}
        response_names = {i["name"] for i in response_data["items"]}

        assert response_data["total"] == len(expected_names)
        assert expected_names == response_names


class TestJobTemplateFiles:
    @pytest.fixture(scope="function")
    async def job_template_data(self, fill_job_template_data, synth_services):
        raw_db_data = await synth_services.crud.template.create(**fill_job_template_data())
        yield JobTemplateBaseDetailedView.model_validate(raw_db_data)

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_TEMPLATES_CREATE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_TEMPLATES_CREATE, Permissions.MAINTAINER]),
        ],
    )
    async def test_upsert_new_file(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_bucket,
        dummy_template,
        make_dummy_file,
        is_owner,
        permissions,
    ):
        parent_id = job_template_data.id
        file_type = "ENTRYPOINT"
        dummy_file_path = make_dummy_file("test_template.py.j2", content=dummy_template)

        requester_email = tester_email if is_owner else "another_" + tester_email

        inject_security_header(requester_email, *permissions)
        with open(dummy_file_path, mode="rb") as template_file:
            response = await client.put(
                f"jobbergate/job-script-templates/{parent_id}/upload/template/{file_type}",
                files={"upload_file": (dummy_file_path.name, template_file, "text/plain")},
            )

        # First, check the response from the upload endpoint
        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"
        response_data = response.json()
        assert response_data is not None
        assert response_data["parent_id"] == parent_id
        assert response_data["filename"] == dummy_file_path.name
        assert response_data["file_type"] == file_type

        # Next, check that the object was inserted into the s3 bucket
        s3_object = await synth_bucket.Object(f"job_script_template_files/{parent_id}/{dummy_file_path.name}")
        response = await s3_object.get()
        file_content = await response["Body"].read()
        assert dummy_template == file_content.decode()

        # Finally, see that the file is included in the parent template file list
        inject_security_header(requester_email, Permissions.JOB_TEMPLATES_READ)
        response = await client.get(f"jobbergate/job-script-templates/{parent_id}")
        assert response.status_code == status.HTTP_200_OK, f"Get failed: {response.text}"

        response_data = response.json()

        template_files = response_data["template_files"]
        assert len(template_files) == 1
        template_file = template_files.pop()
        assert template_file["parent_id"] == parent_id
        assert template_file["filename"] == dummy_file_path.name
        assert template_file["file_type"] == file_type

    @pytest.mark.parametrize(
        "is_owner, permissions, protocol",
        [
            (True, [Permissions.JOB_TEMPLATES_CREATE], "http"),
            (False, [Permissions.ADMIN], "http"),
            (False, [Permissions.JOB_TEMPLATES_CREATE, Permissions.MAINTAINER], "http"),
            (True, [Permissions.JOB_TEMPLATES_CREATE], "https"),
            (False, [Permissions.ADMIN], "https"),
            (False, [Permissions.JOB_TEMPLATES_CREATE, Permissions.MAINTAINER], "https"),
        ],
    )
    async def test_upsert_new_file_by_url__http_https(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_bucket,
        dummy_template,
        is_owner,
        permissions,
        protocol,
        respx_mock,
    ):
        parent_id = job_template_data.id
        file_type = "ENTRYPOINT"

        file_url = f"{protocol}://dummy-domain.com/dummy-file.py.j2"
        respx_mock.get(file_url).mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                content=dummy_template.encode(),
            ),
        )

        requester_email = tester_email if is_owner else "another_" + tester_email

        inject_security_header(requester_email, *permissions)
        response = await client.put(
            f"jobbergate/job-script-templates/{parent_id}/upload-by-url/template/{file_type}",
            params={"file_url": file_url},
        )

        # First, check the response from the upload endpoint
        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"
        response_data = response.json()
        assert response_data is not None
        assert response_data["parent_id"] == parent_id
        assert response_data["filename"] == "dummy-file.py.j2"
        assert response_data["file_type"] == file_type

        # Next, check that the object was inserted into the s3 bucket
        s3_object = await synth_bucket.Object(f"job_script_template_files/{parent_id}/dummy-file.py.j2")
        response = await s3_object.get()
        file_content = await response["Body"].read()
        assert dummy_template == file_content.decode()

        # Finally, see that the file is included in the parent template file list
        inject_security_header(requester_email, Permissions.JOB_TEMPLATES_READ)
        response = await client.get(f"jobbergate/job-script-templates/{parent_id}")
        assert response.status_code == status.HTTP_200_OK, f"Get failed: {response.text}"

        response_data = response.json()

        template_files = response_data["template_files"]
        assert len(template_files) == 1
        template_file = template_files.pop()
        assert template_file["parent_id"] == parent_id
        assert template_file["filename"] == "dummy-file.py.j2"
        assert template_file["file_type"] == file_type

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_TEMPLATES_CREATE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_TEMPLATES_CREATE, Permissions.MAINTAINER]),
        ],
    )
    async def test_upsert_new_file_by_url__s3(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_bucket,
        dummy_template,
        is_owner,
        permissions,
        respx_mock,
    ):
        parent_id = job_template_data.id
        file_type = "ENTRYPOINT"

        s3_url = "s3://dummy-bucket/dummy-file.py.j2"
        https_url = "https://dummy-bucket.s3.amazonaws.com/dummy-file.py.j2"
        respx_mock.get(https_url).mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                content=dummy_template.encode(),
            ),
        )

        requester_email = tester_email if is_owner else "another_" + tester_email

        inject_security_header(requester_email, *permissions)
        response = await client.put(
            f"jobbergate/job-script-templates/{parent_id}/upload-by-url/template/{file_type}",
            params={"file_url": s3_url},
        )

        # First, check the response from the upload endpoint
        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"
        response_data = response.json()
        assert response_data is not None
        assert response_data["parent_id"] == parent_id
        assert response_data["filename"] == "dummy-file.py.j2"
        assert response_data["file_type"] == file_type

        # Next, check that the object was inserted into the s3 bucket
        s3_object = await synth_bucket.Object(f"job_script_template_files/{parent_id}/dummy-file.py.j2")
        response = await s3_object.get()
        file_content = await response["Body"].read()
        assert dummy_template == file_content.decode()

        # Finally, see that the file is included in the parent template file list
        inject_security_header(requester_email, Permissions.JOB_TEMPLATES_READ)
        response = await client.get(f"jobbergate/job-script-templates/{parent_id}")
        assert response.status_code == status.HTTP_200_OK, f"Get failed: {response.text}"

        response_data = response.json()

        template_files = response_data["template_files"]
        assert len(template_files) == 1
        template_file = template_files.pop()
        assert template_file["parent_id"] == parent_id
        assert template_file["filename"] == "dummy-file.py.j2"
        assert template_file["file_type"] == file_type

    @pytest.mark.parametrize("url_path", ["/blah/", "/", ""])
    async def test_upsert_new_file_by_url__fails_on_empty_filename_in_url(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        dummy_template,
        url_path,
        respx_mock,
    ):
        parent_id = job_template_data.id
        file_type = "ENTRYPOINT"

        file_url = "https://dummy-domain.com{url_path}"
        respx_mock.get(file_url).mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                content=dummy_template.encode(),
            ),
        )

        requester_email = tester_email

        inject_security_header(requester_email, Permissions.JOB_TEMPLATES_CREATE)
        response = await client.put(
            f"jobbergate/job-script-templates/{parent_id}/upload-by-url/template/{file_type}",
            params={"file_url": file_url},
        )

        assert response.status_code == status.HTTP_400_BAD_REQUEST
        assert "Filename could not be extracted" in response.text

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_TEMPLATES_CREATE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_TEMPLATES_CREATE, Permissions.MAINTAINER]),
        ],
    )
    async def test_upsert_replace_content(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        make_dummy_file,
        synth_services,
        is_owner,
        permissions,
    ):
        parent_id = job_template_data.id
        file_type = "ENTRYPOINT"
        job_template_filename = "entrypoint.py.j2"

        await synth_services.file.template.upsert(
            parent_id=parent_id,
            filename=job_template_filename,
            upload_content="Old {{ name }}!",
            file_type=file_type,
        )

        new_content = "New {{ name }}!"
        dummy_file_path = make_dummy_file(job_template_filename, content=new_content)

        requester_email = tester_email if is_owner else "another_" + tester_email

        inject_security_header(requester_email, *permissions)
        with open(dummy_file_path, mode="rb") as file:
            response = await client.put(
                f"jobbergate/job-script-templates/{parent_id}/upload/template/{file_type}",
                files={"upload_file": (dummy_file_path.name, file, "text/plain")},
            )

        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"

        template_file = await synth_services.file.template.get(
            parent_id=parent_id, filename=job_template_filename
        )

        assert template_file is not None
        assert template_file.parent_id == parent_id
        assert template_file.filename == job_template_filename
        assert template_file.file_type == file_type

        file_content = await synth_services.file.template.get_file_content(template_file)
        assert file_content.decode() == new_content

    async def test_upsert_rename_file(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_services,
    ):
        parent_id = job_template_data.id
        file_type = "ENTRYPOINT"
        job_template_filename = "entrypoint.py.j2"

        await synth_services.file.template.upsert(
            parent_id=parent_id,
            filename=job_template_filename,
            upload_content="Original {{ name }}!",
            file_type=file_type,
        )

        new_filename = "new_entrypoint.py.j2"

        inject_security_header(tester_email, Permissions.JOB_TEMPLATES_CREATE)
        response = await client.put(
            f"jobbergate/job-script-templates/{parent_id}/upload/template/{file_type}",
            params={"filename": new_filename, "previous_filename": job_template_filename},
        )

        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"

        template_file = await synth_services.file.template.get(parent_id=parent_id, filename=new_filename)

        assert template_file is not None
        assert template_file.parent_id == parent_id
        assert template_file.filename == new_filename
        assert template_file.file_type == file_type

        file_content = await synth_services.file.template.get_file_content(template_file)
        assert file_content.decode() == "Original {{ name }}!"

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_TEMPLATES_CREATE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_TEMPLATES_CREATE, Permissions.MAINTAINER]),
        ],
    )
    async def test_upsert_replace_content_and_rename(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        make_dummy_file,
        synth_services,
        is_owner,
        permissions,
    ):
        parent_id = job_template_data.id
        file_type = "ENTRYPOINT"
        job_template_filename = "entrypoint.py.j2"

        await synth_services.file.template.upsert(
            parent_id=parent_id,
            filename=job_template_filename,
            upload_content="Old {{ name }}!",
            file_type=file_type,
        )

        new_content = "New {{ name }}!"
        new_filename = "new_entrypoint.py.j2"

        dummy_file_path = make_dummy_file(new_filename, content=new_content)
        requester_email = tester_email if is_owner else "another_" + tester_email
        inject_security_header(requester_email, *permissions)
        with open(dummy_file_path, mode="rb") as file:
            response = await client.put(
                f"jobbergate/job-script-templates/{parent_id}/upload/template/{file_type}",
                files={"upload_file": (dummy_file_path.name, file, "text/plain")},
                params={"filename": dummy_file_path.name, "previous_filename": job_template_filename},
            )

        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"

        template_file = await synth_services.file.template.get(parent_id=parent_id, filename=new_filename)

        assert template_file is not None
        assert template_file.parent_id == parent_id
        assert template_file.filename == new_filename
        assert template_file.file_type == file_type

        file_content = await synth_services.file.template.get_file_content(template_file)
        assert file_content.decode() == new_content

    async def test_create__fail_forbidden(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        dummy_template,
        make_dummy_file,
    ):
        parent_id = job_template_data.id
        file_type = "ENTRYPOINT"
        dummy_file_path = make_dummy_file("test_template.py.j2", content=dummy_template)

        owner_email = tester_email
        requester_email = "another_" + owner_email

        inject_security_header(requester_email, Permissions.JOB_TEMPLATES_CREATE)
        with open(dummy_file_path, mode="rb") as template_file:
            response = await client.put(
                f"jobbergate/job-script-templates/{parent_id}/upload/template/{file_type}",
                files={"upload_file": (dummy_file_path.name, template_file, "text/plain")},
            )

        assert response.status_code == status.HTTP_403_FORBIDDEN

    @pytest.mark.parametrize("permission", (Permissions.ADMIN, Permissions.JOB_TEMPLATES_READ))
    async def test_get__success(
        self,
        permission,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_services,
    ):
        parent_id = job_template_data.id
        await synth_services.file.template.upsert(
            parent_id=parent_id,
            filename="test_template.py.j2",
            upload_content="dummy file data",
            file_type="ENTRYPOINT",
        )

        inject_security_header(tester_email, permission)
        response = await client.get(
            f"jobbergate/job-script-templates/{job_template_data.id}/upload/template/test_template.py.j2"
        )

        assert response.status_code == status.HTTP_200_OK, f"Get failed: {response.text}"
        assert response.content.decode() == "dummy file data"

    async def test_get__large_file_success(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_services,
    ):
        """
        Ensure that large files can be retrieved with no problem.
        This was created after strange issues were identified running on FastAPI 0.111.
        """
        large_string = "print(1)\n" * 5000

        parent_id = job_template_data.id
        await synth_services.file.template.upsert(
            parent_id=parent_id,
            filename="test_template.py.j2",
            upload_content=large_string,
            file_type="ENTRYPOINT",
        )

        inject_security_header(tester_email, Permissions.JOB_TEMPLATES_READ)
        response = await client.get(
            f"jobbergate/job-script-templates/{job_template_data.id}/upload/template/test_template.py.j2"
        )

        assert response.status_code == status.HTTP_200_OK, f"Get failed: {response.text}"
        assert response.content.decode() == large_string

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_TEMPLATES_DELETE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_TEMPLATES_DELETE, Permissions.MAINTAINER]),
        ],
    )
    async def test_delete__success(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_bucket,
        synth_services,
        is_owner,
        permissions,
    ):
        parent_id = job_template_data.id
        await synth_services.file.template.upsert(
            parent_id=parent_id,
            filename="test_template.py.j2",
            upload_content="dummy file data",
            file_type="ENTRYPOINT",
        )

        requester_email = tester_email if is_owner else "another_" + tester_email

        inject_security_header(requester_email, *permissions)
        response = await client.delete(
            f"jobbergate/job-script-templates/{parent_id}/upload/template/test_template.py.j2"
        )
        assert response.status_code == status.HTTP_200_OK, f"Delete failed: {response.text}"

        s3_object = await synth_bucket.Object(f"job_script_template_files/{parent_id}/test_template.py.j2")
        with pytest.raises(synth_bucket.meta.client.exceptions.NoSuchKey):
            await s3_object.get()

    async def test_delete__fail_forbidden(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_services,
    ):
        parent_id = job_template_data.id
        await synth_services.file.template.upsert(
            parent_id=parent_id,
            filename="test_template.py.j2",
            upload_content="dummy file data",
            file_type="ENTRYPOINT",
        )

        owner_email = tester_email
        requester_email = "another_" + owner_email

        inject_security_header(requester_email, Permissions.JOB_TEMPLATES_DELETE)
        response = await client.delete(
            f"jobbergate/job-script-templates/{parent_id}/upload/template/test_template.py.j2"
        )
        assert response.status_code == status.HTTP_403_FORBIDDEN


class TestJobTemplateWorkflowFile:
    @pytest.fixture(scope="function")
    async def job_template_data(self, fill_job_template_data, synth_services):
        raw_db_data = await synth_services.crud.template.create(**fill_job_template_data())
        yield JobTemplateListView.model_validate(raw_db_data)

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_TEMPLATES_CREATE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_TEMPLATES_CREATE, Permissions.MAINTAINER]),
        ],
    )
    async def test_create__success(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_bucket,
        dummy_application_source_file,
        make_dummy_file,
        is_owner,
        permissions,
    ):
        parent_id = job_template_data.id
        dummy_file_path = make_dummy_file("test_template.py.j2", content=dummy_application_source_file)
        runtime_config = {"foo": "bar"}

        requester_email = tester_email if is_owner else "another_" + tester_email

        inject_security_header(requester_email, *permissions)
        with open(dummy_file_path, mode="rb") as workflow_file:
            response = await client.put(
                f"jobbergate/job-script-templates/{parent_id}/upload/workflow",
                files={"upload_file": (dummy_file_path.name, workflow_file, "text/plain")},
                data={"runtime_config": json.dumps(runtime_config)},
            )

        # First, check the response from the upload endpoint
        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"
        response_data = response.json()
        assert response_data is not None
        assert response_data["parent_id"] == parent_id
        assert response_data["filename"] == WORKFLOW_FILE_NAME
        assert response_data["runtime_config"] == runtime_config

        # Next, check that the object was inserted into the s3 bucket
        s3_object = await synth_bucket.Object(f"workflow_files/{parent_id}/{WORKFLOW_FILE_NAME}")
        response = await s3_object.get()
        file_content = await response["Body"].read()
        assert dummy_application_source_file == file_content.decode()

        # Finally, see that the file is included in the parent template file list
        inject_security_header(requester_email, Permissions.JOB_TEMPLATES_READ)
        response = await client.get(f"jobbergate/job-script-templates/{parent_id}")
        assert response.status_code == status.HTTP_200_OK, f"Get failed: {response.text}"

        response_data = response.json()

        workflow_files = response_data["workflow_files"]
        assert len(workflow_files) == 1
        workflow_file = workflow_files.pop()
        assert workflow_file["parent_id"] == parent_id
        assert workflow_file["filename"] == WORKFLOW_FILE_NAME
        assert workflow_file["runtime_config"] == runtime_config

    @pytest.mark.parametrize(
        "is_owner, permissions, protocol",
        [
            (True, [Permissions.JOB_TEMPLATES_CREATE], "http"),
            (False, [Permissions.ADMIN], "http"),
            (False, [Permissions.JOB_TEMPLATES_CREATE, Permissions.MAINTAINER], "http"),
            (True, [Permissions.JOB_TEMPLATES_CREATE], "https"),
            (False, [Permissions.ADMIN], "https"),
            (False, [Permissions.JOB_TEMPLATES_CREATE, Permissions.MAINTAINER], "https"),
        ],
    )
    async def test_create_by_url__http_httpx_success(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_bucket,
        dummy_application_source_file,
        is_owner,
        permissions,
        protocol,
        respx_mock,
    ):
        parent_id = job_template_data.id
        runtime_config = {"foo": "bar"}

        file_url = f"{protocol}://dummy-domain.com/dummy-file.py.j2"
        respx_mock.get(file_url).mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                content=dummy_application_source_file.encode(),
            ),
        )

        requester_email = tester_email if is_owner else "another_" + tester_email

        inject_security_header(requester_email, *permissions)
        response = await client.put(
            f"jobbergate/job-script-templates/{parent_id}/upload-by-url/workflow",
            params={"file_url": file_url},
            data=json.dumps(runtime_config),
        )

        # First, check the response from the upload endpoint
        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"
        response_data = response.json()
        assert response_data is not None
        assert response_data["parent_id"] == parent_id
        assert response_data["filename"] == WORKFLOW_FILE_NAME
        assert response_data["runtime_config"] == runtime_config

        # Next, check that the object was inserted into the s3 bucket
        s3_object = await synth_bucket.Object(f"workflow_files/{parent_id}/{WORKFLOW_FILE_NAME}")
        response = await s3_object.get()
        file_content = await response["Body"].read()
        assert dummy_application_source_file == file_content.decode()

        # Finally, see that the file is included in the parent template file list
        inject_security_header(requester_email, Permissions.JOB_TEMPLATES_READ)
        response = await client.get(f"jobbergate/job-script-templates/{parent_id}")
        assert response.status_code == status.HTTP_200_OK, f"Get failed: {response.text}"

        response_data = response.json()

        workflow_files = response_data["workflow_files"]
        assert len(workflow_files) == 1
        workflow_file = workflow_files.pop()
        assert workflow_file["parent_id"] == parent_id
        assert workflow_file["filename"] == WORKFLOW_FILE_NAME
        assert workflow_file["runtime_config"] == runtime_config

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_TEMPLATES_CREATE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_TEMPLATES_CREATE, Permissions.MAINTAINER]),
        ],
    )
    async def test_create_by_url__s3_success(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_bucket,
        dummy_application_source_file,
        is_owner,
        permissions,
        respx_mock,
    ):
        parent_id = job_template_data.id
        runtime_config = {"foo": "bar"}

        s3_url = "s3://dummy-bucket/dummy-file.py.j2"
        https_url = "https://dummy-bucket.s3.amazonaws.com/dummy-file.py.j2"
        respx_mock.get(https_url).mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                content=dummy_application_source_file.encode(),
            ),
        )

        requester_email = tester_email if is_owner else "another_" + tester_email

        inject_security_header(requester_email, *permissions)
        response = await client.put(
            f"jobbergate/job-script-templates/{parent_id}/upload-by-url/workflow",
            params={"file_url": s3_url},
            data=json.dumps(runtime_config),
        )

        # First, check the response from the upload endpoint
        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"
        response_data = response.json()
        assert response_data is not None
        assert response_data["parent_id"] == parent_id
        assert response_data["filename"] == WORKFLOW_FILE_NAME
        assert response_data["runtime_config"] == runtime_config

        # Next, check that the object was inserted into the s3 bucket
        s3_object = await synth_bucket.Object(f"workflow_files/{parent_id}/{WORKFLOW_FILE_NAME}")
        response = await s3_object.get()
        file_content = await response["Body"].read()
        assert dummy_application_source_file == file_content.decode()

        # Finally, see that the file is included in the parent template file list
        inject_security_header(requester_email, Permissions.JOB_TEMPLATES_READ)
        response = await client.get(f"jobbergate/job-script-templates/{parent_id}")
        assert response.status_code == status.HTTP_200_OK, f"Get failed: {response.text}"

        response_data = response.json()

        workflow_files = response_data["workflow_files"]
        assert len(workflow_files) == 1
        workflow_file = workflow_files.pop()
        assert workflow_file["parent_id"] == parent_id
        assert workflow_file["filename"] == WORKFLOW_FILE_NAME
        assert workflow_file["runtime_config"] == runtime_config

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_TEMPLATES_CREATE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_TEMPLATES_CREATE, Permissions.MAINTAINER]),
        ],
    )
    async def test_update__success(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_services,
        make_dummy_file,
        is_owner,
        permissions,
    ):
        parent_id = job_template_data.id
        original_runtime_config = {"foo": "bar"}
        original_content = "import this"

        await synth_services.file.workflow.upsert(
            parent_id=parent_id,
            filename=WORKFLOW_FILE_NAME,
            upload_content=original_content,
            runtime_config=original_runtime_config,
        )

        new_runtime_config = {"new": "config"}
        new_content = "import that"

        requester_email = tester_email if is_owner else "another_" + tester_email

        inject_security_header(requester_email, *permissions)
        with open(make_dummy_file("test_template.py.j2", content=new_content), mode="rb") as workflow_file:
            response = await client.put(
                f"jobbergate/job-script-templates/{parent_id}/upload/workflow",
                files={"upload_file": (workflow_file.name, workflow_file, "text/plain")},
                data={"runtime_config": json.dumps(new_runtime_config)},
            )

        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"

        workflow_file = await synth_services.file.workflow.get(
            parent_id=parent_id,
            filename=WORKFLOW_FILE_NAME,
        )

        assert workflow_file.runtime_config == new_runtime_config
        assert (await synth_services.file.workflow.get_file_content(workflow_file)) == new_content.encode()

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_TEMPLATES_CREATE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_TEMPLATES_CREATE, Permissions.MAINTAINER]),
        ],
    )
    async def test_update_optional_runtime_config__success(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_services,
        make_dummy_file,
        is_owner,
        permissions,
    ):
        parent_id = job_template_data.id
        original_runtime_config = {"foo": "bar"}
        original_content = "import this"

        await synth_services.file.workflow.upsert(
            parent_id=parent_id,
            filename=WORKFLOW_FILE_NAME,
            upload_content=original_content,
            runtime_config=original_runtime_config,
        )

        new_content = "import that"

        requester_email = tester_email if is_owner else "another_" + tester_email

        inject_security_header(requester_email, *permissions)
        with open(make_dummy_file("test_template.py.j2", content=new_content), mode="rb") as workflow_file:
            response = await client.put(
                f"jobbergate/job-script-templates/{parent_id}/upload/workflow",
                files={"upload_file": (workflow_file.name, workflow_file, "text/plain")},
            )

        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"

        workflow_file = await synth_services.file.workflow.get(
            parent_id=parent_id,
            filename=WORKFLOW_FILE_NAME,
        )

        assert workflow_file.runtime_config == original_runtime_config
        assert (await synth_services.file.workflow.get_file_content(workflow_file)) == new_content.encode()

    async def test_update_optional_runtime_config__fail_on_creation_time(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_services,
        make_dummy_file,
    ):
        parent_id = job_template_data.id

        new_content = "import that"

        inject_security_header(tester_email, Permissions.JOB_TEMPLATES_CREATE)
        with open(make_dummy_file("test_template.py.j2", content=new_content), mode="rb") as workflow_file:
            response = await client.put(
                f"jobbergate/job-script-templates/{parent_id}/upload/workflow",
                files={"upload_file": (workflow_file.name, workflow_file, "text/plain")},
            )

        assert response.status_code == status.HTTP_400_BAD_REQUEST
        assert (
            response.json()["detail"]
            == "Runtime configuration is required when the workflow file does not exist"
        )

        with pytest.raises(ServiceError, match="workflow_files row not found"):
            await synth_services.file.workflow.get(parent_id=parent_id, filename=WORKFLOW_FILE_NAME)

    async def test_create__fail_forbidden(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        dummy_application_source_file,
        make_dummy_file,
    ):
        parent_id = job_template_data.id
        dummy_file_path = make_dummy_file("test_template.py.j2", content=dummy_application_source_file)
        runtime_config = {"foo": "bar"}

        owner_email = tester_email
        requester_email = "another_" + owner_email

        inject_security_header(requester_email, Permissions.JOB_TEMPLATES_CREATE)
        with open(dummy_file_path, mode="rb") as workflow_file:
            response = await client.put(
                f"jobbergate/job-script-templates/{parent_id}/upload/workflow",
                files={"upload_file": (dummy_file_path.name, workflow_file, "text/plain")},
                data={"runtime_config": json.dumps(runtime_config)},
            )

        # First, check the response from the upload endpoint
        assert response.status_code == status.HTTP_403_FORBIDDEN, f"Upload failed: {response.text}"

    @pytest.mark.parametrize("permission", (Permissions.ADMIN, Permissions.JOB_TEMPLATES_READ))
    async def test_get__success(
        self, permission, client, tester_email, inject_security_header, job_template_data, synth_services
    ):
        parent_id = job_template_data.id
        await synth_services.file.workflow.upsert(
            parent_id=parent_id,
            filename=WORKFLOW_FILE_NAME,
            upload_content="import this",
            runtime_config={"foo": "bar"},
        )

        inject_security_header(tester_email, permission)
        response = await client.get(f"jobbergate/job-script-templates/{parent_id}/upload/workflow")

        assert response.status_code == status.HTTP_200_OK, f"Get failed: {response.text}"
        assert response.content.decode() == "import this"

    async def test_get__large_file_success(
        self, client, tester_email, inject_security_header, job_template_data, synth_services
    ):
        """
        Ensure that large files can be retrieved with no problem.
        This was created after strange issues were identified running on FastAPI 0.111.
        """
        large_string = "print(1)\n" * 5000

        parent_id = job_template_data.id
        await synth_services.file.workflow.upsert(
            parent_id=parent_id,
            filename=WORKFLOW_FILE_NAME,
            upload_content=large_string,
            runtime_config={"foo": "bar"},
        )

        inject_security_header(tester_email, Permissions.JOB_TEMPLATES_READ)
        response = await client.get(f"jobbergate/job-script-templates/{parent_id}/upload/workflow")

        assert response.status_code == status.HTTP_200_OK, f"Get failed: {response.text}"
        assert response.content.decode() == large_string

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_TEMPLATES_DELETE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_TEMPLATES_DELETE, Permissions.MAINTAINER]),
        ],
    )
    async def test_delete__success(
        self,
        client,
        tester_email,
        inject_security_header,
        job_template_data,
        synth_bucket,
        synth_services,
        is_owner,
        permissions,
    ):
        parent_id = job_template_data.id
        upserted_instance = await synth_services.file.workflow.upsert(
            parent_id=parent_id,
            filename=WORKFLOW_FILE_NAME,
            upload_content="import this",
            runtime_config={"foo": "bar"},
        )

        requester_email = tester_email if is_owner else "another_" + tester_email

        inject_security_header(requester_email, *permissions)
        response = await client.delete(f"jobbergate/job-script-templates/{parent_id}/upload/workflow")
        assert response.status_code == status.HTTP_200_OK, f"Delete failed: {response.text}"

        s3_object = await synth_bucket.Object(upserted_instance.file_key)
        with pytest.raises(synth_bucket.meta.client.exceptions.NoSuchKey):
            await s3_object.get()



import pytest
from pydantic import ValidationError

from jobbergate_api.apps.job_script_templates.schemas import JobTemplateCreateRequest


def test_not_empty_str__validator_not_applied_to_non_empty_strings():
    request = JobTemplateCreateRequest(name="test-name")

    assert request.name == "test-name"


def test_not_empty_str__validator_applied_to_empty_strings():
    with pytest.raises(ValidationError, match="Cannot be an empty string"):
        JobTemplateCreateRequest(name="")


def test_empty_str_to_none__validator_not_applied_to_non_empty_strings():
    request = JobTemplateCreateRequest(
        name="test-name",
        identifier="test-identifier",
    )

    assert request.identifier == "test-identifier"


def test_empty_str_to_none__validator_applied_to_empty_strings():
    request = JobTemplateCreateRequest(
        name="test-name",
        identifier="",
    )

    assert request.identifier is None



"""
RabbitMQ notification system for Jobbergate.
"""

import asyncio
import json
import socket
from contextlib import asynccontextmanager
from typing import Optional

import aio_pika
from loguru import logger

from jobbergate_api.apps.job_submissions.models import JobSubmission
from jobbergate_api.config import settings
from jobbergate_api.retry_utils import async_retry


@asynccontextmanager
async def rabbitmq_connect(
    exchange_name=None,
    do_purge=False,
):
    """
    Connect to a RabbitMQ queue and exchange.
    """
    if exchange_name is None:
        exchange_name = settings.RABBITMQ_DEFAULT_EXCHANGE

    event_loop = asyncio.get_running_loop()
    connection = await aio_pika.connect_robust(
        host=settings.RABBITMQ_HOST,
        login=settings.RABBITMQ_USERNAME,
        password=settings.RABBITMQ_PASSWORD,
        client_properties={"connection_name": socket.gethostname()},
        loop=event_loop,
    )
    channel = await connection.channel(publisher_confirms=True)
    exchange = await channel.declare_exchange(
        name=exchange_name,
        type=aio_pika.ExchangeType.DIRECT,
        durable=True,
    )
    queue_name = "jobs"
    routing_key = "status"
    declared_queue = await channel.declare_queue(name=queue_name, durable=True)
    await declared_queue.bind(exchange, routing_key)

    try:
        yield (exchange, declared_queue)

    finally:
        if do_purge:
            await declared_queue.purge(timeout=1)
        await declared_queue.unbind(exchange, routing_key)
        await connection.close()


async def _publish_message(
    exchange: aio_pika.Exchange,
    message_payload: dict,
    organization_id: Optional[str],
) -> None:
    """Publish a single message to RabbitMQ exchange."""
    await exchange.publish(
        message=aio_pika.Message(
            body=json.dumps(message_payload).encode("utf-8"),
            delivery_mode=aio_pika.DeliveryMode.PERSISTENT,
            headers={"organization": organization_id},
        ),
        routing_key="status",
    )


async def publish_status_change(
    job_submission: JobSubmission,
    organization_id: Optional[str] = None,
) -> bool:
    """
    Publish a status change for a JobSubmission to the RabbitMQ exchange used for notifications.

    Returns True if successful, False if failed after max retries.
    """
    if settings.RABBITMQ_HOST is None:
        return True  # Skip when RabbitMQ is not configured

    logger.debug("Publishing status change to notification queue")

    message_payload = {
        "path": f"jobs.job_submissions.{job_submission.id}",
        "user_email": job_submission.owner_email,
        "action": "status",
        "additional_context": {
            "status": job_submission.status,
            "slurm_job_state": job_submission.slurm_job_state,
        },
    }

    async def _publish_with_connection():
        async with rabbitmq_connect(exchange_name=organization_id) as (exchange, _):
            await _publish_message(exchange, message_payload, organization_id)

    def on_retry_error(exc: Exception, attempt: int) -> None:
        logger.warning(f"Failed to publish status change notification (attempt {attempt}): {exc}")

    result = await async_retry(
        _publish_with_connection,
        max_attempts=3,
        initial_delay=1.0,
        backoff_factor=2.0,
        on_error=on_retry_error,
    )

    if result is None:
        logger.error("Failed to publish status change notification after 3 retry attempts")
        return False

    return True



"""Make parent job-script null on delete

Revision ID: b15213c159f1
Revises: 3bc095975b53
Create Date: 2023-09-06 18:15:40.385837

"""

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision = "b15213c159f1"
down_revision = "3bc095975b53"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_constraint("job_submissions_job_script_id_fkey", "job_submissions", type_="foreignkey")
    op.create_foreign_key(
        None, "job_submissions", "job_scripts", ["job_script_id"], ["id"], ondelete="SET NULL"
    )


def downgrade():
    op.drop_constraint(None, "job_submissions", type_="foreignkey")
    op.create_foreign_key(
        "job_submissions_job_script_id_fkey", "job_submissions", "job_scripts", ["job_script_id"], ["id"]
    )



"""
Provide helper commands for local development.
"""

import typer

from dev_tools import db, dev_server, show_env, metrics

app = typer.Typer()
app.command(name="dev-server")(dev_server.dev_server)
app.command(name="show-env")(show_env.show_env)
app.command(name="generate-metrics")(metrics.generate_metrics)
app.add_typer(db.app, name="db")


if __name__ == "__main__":
    app()



"""
Provide command for starting a local development server for the API.
"""

from time import sleep

import typer
import uvicorn
from loguru import logger
from sqlalchemy import create_engine, text

from jobbergate_api.storage import build_db_url


def _wait_for_db(wait_count, wait_interval):
    """
    Wait for a database connection.

    Loop for a maximum of ``wait_count`` times where each lasts for ``wait_interval`` seconds. If the
    connection resolves before the time is up, return normally. If the database fails to connect, raise a
    ``RuntimeError``.
    """
    database_url = build_db_url()
    logger.debug(f"database url is: {build_db_url()}")
    count = 0
    while count < wait_count:
        logger.debug(f"Checking health of database at {database_url}: Attempt #{count}")
        count += 1
        try:
            engine = create_engine(database_url)
            with engine.connect() as db:
                db.execute(text("select version()"))
            return
        except Exception as err:
            logger.warning(f"Database is not yet healthy: {err}")
        sleep(wait_interval)

    raise RuntimeError("Could not connect to the database")


def dev_server(
    db_wait_count: int = typer.Option(3, help="How many times to attempt a check"),
    db_wait_interval: float = typer.Option(5.0, help="Seconds to wait between checks"),
    port: int = typer.Option(5000, help="The port where the server should listen"),
    log_level: str = typer.Option("DEBUG", help="The level to log uvicorn output"),
):
    """
    Start a development server locally.
    """
    try:
        logger.info("Waiting for the database")
        _wait_for_db(db_wait_count, db_wait_interval)
    except Exception:
        logger.error("Database is not available")
        typer.Exit(1)

    uvicorn.run(
        "jobbergate_api.main:app",
        host="127.0.0.1",
        port=port,
        reload=True,
        log_level=log_level.lower(),
    )



"""add job_progress table for tracking job submission state

Revision ID: 31382ad313d1
Revises: 99c3877d0f10
Create Date: 2025-03-25 17:38:11.047745

"""
import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision = "31382ad313d1"
down_revision = "99c3877d0f10"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "job_progresses",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("job_submission_id", sa.Integer(), nullable=False),
        sa.Column("timestamp", sa.DateTime(timezone=True), nullable=False),
        sa.Column("slurm_job_state", sa.String(), nullable=True),
        sa.Column("additional_info", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["job_submission_id"], ["job_submissions.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("job_progresses")
    # ### end Alembic commands ###



"""Tests for the /job-scripts/ endpoint."""

import httpx
import pytest
from fastapi import status

from jobbergate_api.apps.permissions import Permissions

# Not using the synth_session fixture in a route that needs the database is unsafe
pytest.mark.usefixtures("synth_session")


@pytest.mark.parametrize("permission", (Permissions.ADMIN, Permissions.JOB_SCRIPTS_CREATE))
async def test_create_stand_alone_job_script__success(
    client, permission, fill_job_script_data, inject_security_header, synth_services
):
    """Test a stand alone job script can be create."""
    payload = fill_job_script_data()

    tester_email = payload.pop("owner_email")
    inject_security_header(tester_email, permission)

    response = await client.post("jobbergate/job-scripts", json=payload)
    assert response.status_code == 201, f"Create failed: {response.text}"
    response_data = response.json()

    assert response_data["name"] == payload["name"]
    assert response_data["description"] == payload["description"]
    assert response_data["owner_email"] == tester_email
    assert "files" not in response_data
    assert response_data["parent_template_id"] is None

    created_id = response_data["id"]

    # Make sure the data was actually inserted into the database
    assert (await synth_services.crud.job_script.count()) == 1
    instance = await synth_services.crud.job_script.get(created_id)
    assert instance is not None
    assert instance.id == created_id
    assert instance.name == payload["name"]
    assert instance.description == payload["description"]
    assert instance.parent_template_id is None

    # Make sure that the data can be retrieved with a GET request
    inject_security_header(tester_email, Permissions.JOB_SCRIPTS_READ)
    response = await client.get(f"jobbergate/job-scripts/{created_id}")
    assert response.status_code == 200
    response_data = response.json()
    assert response_data["name"] == payload["name"]
    assert response_data["description"] == payload["description"]
    assert response_data["owner_email"] == tester_email
    assert response_data["files"] == []
    assert response_data["parent_template_id"] is None


@pytest.mark.parametrize("permission", (Permissions.ADMIN, Permissions.JOB_SCRIPTS_CREATE))
async def test_clone_job_script__success(
    client, permission, fill_job_script_data, inject_security_header, tester_email, synth_services
):
    original_instance = await synth_services.crud.job_script.create(
        **fill_job_script_data(owner_email=tester_email)
    )
    parent_id = original_instance.id
    await synth_services.file.job_script.upsert(
        parent_id=parent_id,
        filename="entrypoint.py",
        upload_content="print('dummy file data')",
        file_type="ENTRYPOINT",
    )
    await synth_services.file.job_script.upsert(
        parent_id=parent_id,
        filename="support.sh",
        upload_content="echo 'dummy file data'",
        file_type="SUPPORT",
    )

    new_owner_email = "new_" + tester_email

    inject_security_header(new_owner_email, permission)
    response = await client.post(f"jobbergate/job-scripts/clone/{original_instance.id}")

    assert response.status_code == 201, f"Clone failed: {response.text}"
    response_data = response.json()

    cloned_id = response_data["id"]

    assert cloned_id != original_instance.id
    assert response_data["name"] == original_instance.name
    assert response_data["description"] == original_instance.description
    assert response_data["owner_email"] == new_owner_email
    assert response_data["cloned_from_id"] == original_instance.id

    assert {f["filename"] for f in response_data["files"]} == {"entrypoint.py", "support.sh"}


async def test_clone_job_script__replace_base_values(
    client, fill_job_script_data, inject_security_header, tester_email, synth_services
):
    original_instance = await synth_services.crud.job_script.create(
        **fill_job_script_data(owner_email=tester_email)
    )

    new_owner_email = "new_" + tester_email

    payload = {
        "name": "new name",
        "description": "new description",
    }

    inject_security_header(new_owner_email, Permissions.JOB_SCRIPTS_CREATE)
    response = await client.post(f"jobbergate/job-scripts/clone/{original_instance.id}", json=payload)

    assert response.status_code == 201, f"Clone failed: {response.text}"
    response_data = response.json()

    assert response_data["name"] == payload["name"]
    assert response_data["description"] == payload["description"]
    assert response_data["owner_email"] == new_owner_email
    assert response_data["cloned_from_id"] == original_instance.id


async def test_clone_job_script__fail_unauthorized(client, fill_job_script_data, synth_services):
    original_instance = await synth_services.crud.job_script.create(**fill_job_script_data())

    response = await client.post(f"jobbergate/job-scripts/clone/{original_instance.id}")

    assert response.status_code == 401


async def test_clone_job_script__fail_not_found(
    client,
    inject_security_header,
    tester_email,
    synth_services,
):
    assert (await synth_services.crud.job_script.count()) == 0
    inject_security_header(tester_email, Permissions.JOB_SCRIPTS_CREATE)
    response = await client.post("jobbergate/job-scripts/clone/0")

    assert response.status_code == 404


@pytest.mark.parametrize("permission", (Permissions.ADMIN, Permissions.JOB_SCRIPTS_CREATE))
async def test_render_job_script_from_template__success(
    permission,
    fill_job_template_data,
    fill_job_script_data,
    client,
    inject_security_header,
    dummy_template,
    tester_email,
    job_script_data_as_string,
    synth_services,
):
    """
    Test POST /job_scripts/render-from-template correctly creates a job_script.

    This test proves that a job_script is successfully created via a POST request to the /job-scripts/
    endpoint. We show this by asserting that the job_script is created in the database after the post
    request is made, the correct status code (201) is returned.
    """
    base_template = await synth_services.crud.template.create(**fill_job_template_data())

    template_name = "entrypoint.sh.j2"
    job_script_name = template_name.removesuffix(".j2")
    await synth_services.file.template.upsert(
        parent_id=base_template.id,
        file_type="ENTRYPOINT",
        filename=template_name,
        upload_content=dummy_template,
    )

    payload = {
        "create_request": fill_job_script_data(),
        "render_request": {
            "template_output_name_mapping": {template_name: job_script_name},
            "sbatch_params": ["--partition=debug", "--time=00:30:00"],
            "param_dict": {"data": {"job_name": "rats", "partition": "debug"}},
        },
    }

    inject_security_header(tester_email, permission)
    response = await client.post(
        f"/jobbergate/job-scripts/render-from-template/{base_template.id}",
        json=payload,
    )

    assert response.status_code == status.HTTP_201_CREATED, f"Render failed: {response.text}"

    assert (await synth_services.crud.job_script.count()) == 1

    response_data = response.json()

    assert response_data["name"] == payload["create_request"]["name"]
    assert response_data["owner_email"] == tester_email
    assert response_data["description"] == payload["create_request"]["description"]
    assert job_script_name in [f["filename"] for f in response_data["files"]]
    assert response_data["parent_template_id"] == base_template.id

    instance = await synth_services.file.job_script.get(response_data["id"], job_script_name)
    rendered_file_contents = await synth_services.file.job_script.get_file_content(instance)
    assert rendered_file_contents.decode("utf-8") == job_script_data_as_string


async def test_render_job_script_from_template__no_entrypoint(
    fill_job_template_data,
    fill_job_script_data,
    client,
    inject_security_header,
    dummy_template,
    tester_email,
    synth_services,
):
    """
    Test POST /job_scripts/render-from-template raises 400 if no entrypoint is found.
    """
    base_template = await synth_services.crud.template.create(**fill_job_template_data())

    template_name = "entrypoint.sh.j2"
    job_script_name = template_name.removesuffix(".j2")
    await synth_services.file.template.upsert(
        parent_id=base_template.id,
        file_type="SUPPORT",
        filename=template_name,
        upload_content=dummy_template,
    )

    payload = {
        "create_request": fill_job_script_data(),
        "render_request": {
            "template_output_name_mapping": {template_name: job_script_name},
            "param_dict": {"data": {"job_name": "rats", "partition": "debug"}},
        },
    }

    inject_security_header(tester_email, Permissions.JOB_SCRIPTS_CREATE)
    response = await client.post(
        f"/jobbergate/job-scripts/render-from-template/{base_template.id}",
        json=payload,
    )

    assert response.status_code == status.HTTP_400_BAD_REQUEST, f"Render failed: {response.text}"
    assert "Exactly one entrypoint file must be specified, got 0" in response.text


async def test_render_job_script_from_template__multiple_entrypoints(
    fill_job_template_data,
    fill_job_script_data,
    client,
    inject_security_header,
    dummy_template,
    tester_email,
    synth_services,
):
    """
    Test POST /job_scripts/render-from-template raises 400 if more than one entrypoint is found.
    """
    base_template = await synth_services.crud.template.create(**fill_job_template_data())

    template_name_1 = "entrypoint-1.py.j2"
    job_script_name_1 = template_name_1.removesuffix(".j2")
    await synth_services.file.template.upsert(
        parent_id=base_template.id,
        file_type="ENTRYPOINT",
        filename=template_name_1,
        upload_content=dummy_template,
    )
    template_name_2 = "entrypoint-2.py.j2"
    job_script_name_2 = template_name_2.removesuffix(".j2")
    await synth_services.file.template.upsert(
        parent_id=base_template.id,
        file_type="ENTRYPOINT",
        filename=template_name_2,
        upload_content=dummy_template,
    )

    payload = {
        "create_request": fill_job_script_data(),
        "render_request": {
            "template_output_name_mapping": {
                template_name_1: job_script_name_1,
                template_name_2: job_script_name_2,
            },
            "param_dict": {"data": {"job_name": "rats", "partition": "debug"}},
        },
    }

    inject_security_header(tester_email, Permissions.JOB_SCRIPTS_CREATE)
    response = await client.post(
        f"/jobbergate/job-scripts/render-from-template/{base_template.id}",
        json=payload,
    )

    assert response.status_code == status.HTTP_400_BAD_REQUEST, f"Render failed: {response.text}"
    assert "Exactly one entrypoint file must be specified, got 2" in response.text


async def test_render_job_script_from_template__template_file_unavailable(
    fill_job_template_data,
    fill_job_script_data,
    client,
    inject_security_header,
    tester_email,
    synth_services,
):
    """
    Test POST /job_scripts/render-from-template can't create a job_script if the template file is unavailable.
    """
    base_template = await synth_services.crud.template.create(**fill_job_template_data())

    template_name = "entrypoint.sh.j2"
    job_script_name = template_name.removesuffix(".j2")

    payload = {
        "create_request": fill_job_script_data(),
        "render_request": {
            "template_output_name_mapping": {template_name: job_script_name},
            "sbatch_params": ["--partition=debug", "--time=00:30:00"],
            "param_dict": {"data": {"job_name": "rats", "partition": "debug"}},
        },
    }

    inject_security_header(tester_email, Permissions.JOB_SCRIPTS_CREATE)
    response = await client.post(
        f"/jobbergate/job-scripts/render-from-template/{base_template.id}",
        json=payload,
    )

    assert response.status_code == status.HTTP_400_BAD_REQUEST

    assert (await synth_services.crud.job_script.count()) == 0


async def test_render_job_script_from_template__bad_permission(
    fill_job_template_data,
    fill_job_script_data,
    client,
    inject_security_header,
    tester_email,
    synth_services,
):
    """
    Test that it is not possible to create job_script without proper permission.
    """
    base_template = await synth_services.crud.template.create(**fill_job_template_data())

    template_name = "entrypoint.sh.j2"
    job_script_name = template_name.removesuffix(".j2")
    payload = {
        "create_request": fill_job_script_data(),
        "render_request": {
            "template_output_name_mapping": {template_name: job_script_name},
            "sbatch_params": ["--partition=debug", "--time=00:30:00"],
            "param_dict": {"data": {"job_name": "rats", "partition": "debug"}},
        },
    }

    inject_security_header(tester_email, "INVALID_PERMISSION")

    response = await client.post(
        f"/jobbergate/job-scripts/render-from-template/{base_template.id}",
        json=payload,
    )
    assert response.status_code == status.HTTP_403_FORBIDDEN


async def test_render_job_script_from_template__without_template(
    fill_job_script_data,
    client,
    inject_security_header,
    tester_email,
    synth_session,
):
    """
    Test POST /job_scripts/render-from-template can't create a job_script if the template file is unavailable.
    """
    template_name = "entrypoint.sh.j2"
    job_script_name = template_name.removesuffix(".j2")
    payload = {
        "create_request": fill_job_script_data(),
        "render_request": {
            "template_output_name_mapping": {template_name: job_script_name},
            "sbatch_params": ["--partition=debug", "--time=00:30:00"],
            "param_dict": {"data": {"job_name": "rats", "partition": "debug"}},
        },
    }

    inject_security_header(tester_email, Permissions.JOB_SCRIPTS_CREATE)
    response = await client.post(
        "/jobbergate/job-scripts/render-from-template/9999",
        json=payload,
    )
    assert response.status_code == status.HTTP_404_NOT_FOUND


@pytest.mark.parametrize("permission", (Permissions.ADMIN, Permissions.JOB_SCRIPTS_READ))
async def test_get_job_script_by_id__success(
    permission,
    fill_job_script_data,
    client,
    inject_security_header,
    tester_email,
    synth_services,
):
    """
    Test GET /job-scripts/<id>.

    This test proves that GET /job-scripts/<id> returns the correct job-script, owned by
    the user making the request. We show this by asserting that the job_script data
    returned in the response is equal to the job_script data that exists in the database
    for the given job_script id.
    """
    inserted_instance = await synth_services.crud.job_script.create(**fill_job_script_data())

    inject_security_header(tester_email, permission)
    response = await client.get(f"/jobbergate/job-scripts/{inserted_instance.id}")

    assert response.status_code == status.HTTP_200_OK, f"Get failed: {response.text}"

    data = response.json()
    assert data["id"] == inserted_instance.id
    assert data["name"] == inserted_instance.name
    assert data["owner_email"] == inserted_instance.owner_email
    assert data["parent_template_id"] is None
    assert data["files"] == []


async def test_get_job_script_by_id__invalid_id(
    client,
    inject_security_header,
    tester_email,
    synth_session,
):
    """
    Test the correct response code is returned when a job_script does not exist.

    This test proves that GET /job-script/<id> returns the correct response code when the
    requested job_script does not exist. We show this by asserting that the status code
    returned is what we would expect given the job_script requested doesn't exist (404).
    """
    inject_security_header(tester_email, Permissions.JOB_SCRIPTS_READ)
    response = await client.get("/jobbergate/job-scripts/9999")

    assert response.status_code == status.HTTP_404_NOT_FOUND


async def test_get_job_script_by_id__bad_permission(
    fill_job_script_data,
    client,
    inject_security_header,
    tester_email,
    synth_services,
):
    """
    Test GET /job-scripts/<id>.

    This test proves that GET /job-scripts/<id> returns the correct job-script, owned by
    the user making the request. We show this by asserting that the job_script data
    returned in the response is equal to the job_script data that exists in the database
    for the given job_script id.
    """
    inserted_instance = await synth_services.crud.job_script.create(**fill_job_script_data())

    inject_security_header(tester_email, "INVALID_PERMISSION")
    response = await client.get(f"/jobbergate/job-scripts/{inserted_instance.id}")

    assert response.status_code == status.HTTP_403_FORBIDDEN


@pytest.mark.parametrize(
    "is_owner, permissions",
    [
        (True, [Permissions.JOB_SCRIPTS_UPDATE]),
        (False, [Permissions.ADMIN]),
        (False, [Permissions.JOB_SCRIPTS_UPDATE, Permissions.MAINTAINER]),
    ],
)
async def test_update_job_script__success(
    client,
    fill_job_script_data,
    inject_security_header,
    tester_email,
    synth_services,
    is_owner,
    permissions,
):
    instance = await synth_services.crud.job_script.create(**fill_job_script_data())

    requester_email = tester_email if is_owner else "another_" + tester_email

    payload = {
        "name": "new-name",
        "description": "new-description",
    }

    inject_security_header(requester_email, *permissions)
    response = await client.put(f"jobbergate/job-scripts/{instance.id}", json=payload)

    assert response.status_code == 200, f"Update failed: {response.text}"
    response_data = response.json()
    assert response_data["name"] == payload["name"]
    assert response_data["description"] == payload["description"]


async def test_update_job_script__fail_not_found(
    client,
    inject_security_header,
    tester_email,
    synth_services,
):
    payload = {
        "name": "new-name",
        "description": "new-description",
    }
    inject_security_header(tester_email, Permissions.JOB_SCRIPTS_UPDATE)
    response = await client.put("jobbergate/job-scripts/0", json=payload)

    assert response.status_code == 404


async def test_update_job_script__fail_unauthorized(client):
    payload = {
        "name": "new-name",
        "description": "new-description",
    }
    response = await client.put("jobbergate/job-scripts/0", json=payload)

    assert response.status_code == 401


async def test_update_job_script__fail_forbidden(
    client,
    fill_job_script_data,
    inject_security_header,
    tester_email,
    synth_services,
):
    instance = await synth_services.crud.job_script.create(**fill_job_script_data())

    owner_email = tester_email
    requester_email = "another_" + owner_email

    payload = {
        "name": "new-name",
        "description": "new-description",
    }

    inject_security_header(requester_email, Permissions.JOB_SCRIPTS_UPDATE)
    response = await client.put(f"jobbergate/job-scripts/{instance.id}", json=payload)

    assert response.status_code == 403


@pytest.mark.parametrize(
    "is_owner, permissions",
    [
        (True, [Permissions.JOB_SCRIPTS_DELETE]),
        (False, [Permissions.ADMIN]),
        (False, [Permissions.JOB_SCRIPTS_DELETE, Permissions.MAINTAINER]),
    ],
)
async def test_delete_job_script__success(
    client,
    inject_security_header,
    fill_job_script_data,
    synth_services,
    is_owner,
    permissions,
):
    instance = await synth_services.crud.job_script.create(**fill_job_script_data())

    owner_email = instance.owner_email
    requester_email = owner_email if is_owner else "another_" + owner_email

    inject_security_header(requester_email, *permissions)
    response = await client.delete(f"jobbergate/job-scripts/{instance.id}")
    assert response.status_code == 204, f"Delete failed: {response.text}"

    assert (await synth_services.crud.job_script.count()) == 0


async def test_delete_job_script__fail_not_found(
    client,
    inject_security_header,
    tester_email,
    synth_services,
):
    inject_security_header(tester_email, Permissions.JOB_SCRIPTS_DELETE)
    response = await client.delete("jobbergate/job-scripts/0")
    assert response.status_code == 404


async def test_delete_job_script__fail_forbidden(
    client,
    inject_security_header,
    fill_job_script_data,
    synth_services,
):
    instance = await synth_services.crud.job_script.create(**fill_job_script_data())

    owner_email = instance.owner_email
    requester_email = "another_" + owner_email

    inject_security_header(requester_email, Permissions.JOB_SCRIPTS_DELETE)
    response = await client.delete(f"jobbergate/job-scripts/{instance.id}")
    assert response.status_code == 403

    assert (await synth_services.crud.job_script.count()) == 1


class TestListJobScripts:
    """Test the list endpoint."""

    @pytest.fixture(scope="function")
    async def job_scripts_list(self, fill_all_job_script_data, synth_services):
        data = fill_all_job_script_data(
            {
                "name": "name-1",
                "description": "desc-1",
            },
            {
                "name": "name-2",
                "description": "desc-2",
            },
            {
                "name": "name-3",
                "description": "desc-3",
            },
            {
                "name": "name-4",
                "description": "desc-4",
                "owner_email": "test-test@pytest.com",
            },
            {
                "name": "name-5",
                "description": "desc-5",
                "owner_email": "test-test@pytest.com",
                "is_archived": True,
            },
        )
        for item in data:
            await synth_services.crud.job_script.create(**item)
        yield data

    @pytest.mark.parametrize("permission", (Permissions.ADMIN, Permissions.JOB_SCRIPTS_READ))
    async def test_list_job_scripts__all_success(
        self,
        permission,
        client,
        tester_email,
        inject_security_header,
        job_scripts_list,
    ):
        inject_security_header(tester_email, permission)
        response = await client.get("jobbergate/job-scripts?include_archived=True&sort_field=id")
        assert response.status_code == 200, f"Get failed: {response.text}"

        response_data = response.json()
        assert response_data["total"] == len(job_scripts_list)
        assert response_data["page"] == 1
        assert response_data["size"] == 50
        assert response_data["pages"] == 1

        for response_item, expected_item in zip(response_data["items"], job_scripts_list):
            assert response_item["name"] == expected_item["name"]
            assert response_item["description"] == expected_item["description"]
            assert response_item["owner_email"] == expected_item["owner_email"]
            assert response_item["is_archived"] == expected_item["is_archived"]
            assert response_item["template"] is None

    async def test_list_job_scripts__ignore_archived(
        self,
        client,
        tester_email,
        inject_security_header,
        job_scripts_list,
    ):
        inject_security_header(tester_email, Permissions.JOB_SCRIPTS_READ)
        response = await client.get("jobbergate/job-scripts")
        assert response.status_code == 200, f"Get failed: {response.text}"

        response_data = response.json()

        expected_names = {i["name"] for i in job_scripts_list if i["is_archived"] is False}
        response_names = {i["name"] for i in response_data["items"]}

        assert response_data["total"] == len(expected_names)
        assert expected_names == response_names

    async def test_list_job_scripts__user_only(
        self,
        client,
        tester_email,
        inject_security_header,
        job_scripts_list,
    ):
        inject_security_header(tester_email, Permissions.JOB_SCRIPTS_READ)
        response = await client.get("jobbergate/job-scripts?user_only=True&include_archived=True")
        assert response.status_code == 200, f"Get failed: {response.text}"

        response_data = response.json()

        expected_names = {i["name"] for i in job_scripts_list if i["owner_email"] == tester_email}
        actual_names = {i["name"] for i in response_data["items"]}

        assert response_data["total"] == len(expected_names)
        assert expected_names == actual_names

    async def test_list_job_scripts__with_parent(
        self,
        client,
        tester_email,
        inject_security_header,
        fill_job_template_data,
        fill_all_job_script_data,
        synth_services,
        job_script_data_as_string,
    ):
        base_template = await synth_services.crud.template.create(**fill_job_template_data())

        data = fill_all_job_script_data(
            {"name": "name-1", "parent_template_id": base_template.id},
            {"name": "name-2"},
        )

        for item in data:
            job_script_data = await synth_services.crud.job_script.create(**item)

            id = job_script_data.id
            file_type = "ENTRYPOINT"
            job_script_filename = "entrypoint.sh"

            await synth_services.file.job_script.upsert(
                parent_id=id,
                filename=job_script_filename,
                upload_content=job_script_data_as_string,
                file_type=file_type,
            )

        inject_security_header(tester_email, Permissions.JOB_SCRIPTS_READ)
        response = await client.get("jobbergate/job-scripts")
        assert response.status_code == 200, f"Get failed: {response.text}"

        response_data = response.json()

        expected_names = {i["name"] for i in data}
        actual_names = {i["name"] for i in response_data["items"]}

        assert response_data["total"] == len(expected_names)
        assert expected_names == actual_names

        assert response_data["items"][0]["template"] is not None
        assert response_data["items"][0]["template"]["name"] == base_template.name
        assert response_data["items"][1]["template"] is None

    async def test_list_job_scripts__search(
        self,
        client,
        tester_email,
        inject_security_header,
        fill_all_job_script_data,
        synth_services,
    ):
        data = fill_all_job_script_data(
            {
                "name": "instance-one",
                "description": "the first",
                "owner_email": "user1@test.com",
            },
            {
                "name": "item-two",
                "description": "second item",
                "owner_email": "user2@test.com",
            },
            {
                "name": "instance-three",
                "description": "a final instance",
                "owner_email": "final@test.com",
            },
        )

        for item in data:
            await synth_services.crud.job_script.create(**item)

        inject_security_header(tester_email, Permissions.JOB_SCRIPTS_READ)
        response = await client.get("jobbergate/job-scripts", params={"search": "instance"})
        assert response.status_code == 200, f"Get failed: {response.text}"

        response_data = response.json()

        expected_names = {"instance-one", "instance-three"}
        actual_names = {i["name"] for i in response_data["items"]}

        assert response_data["total"] == len(expected_names)
        assert expected_names == actual_names


class TestJobScriptFiles:
    @pytest.fixture(scope="function")
    async def job_script_data(self, fill_job_script_data, synth_services):
        raw_db_data = await synth_services.crud.job_script.create(**fill_job_script_data())
        yield raw_db_data

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_SCRIPTS_CREATE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_SCRIPTS_CREATE, Permissions.MAINTAINER]),
        ],
    )
    async def test_upsert_new_file(
        self,
        client,
        inject_security_header,
        job_script_data,
        job_script_data_as_string,
        make_dummy_file,
        synth_services,
        is_owner,
        permissions,
    ):
        id = job_script_data.id
        file_type = "ENTRYPOINT"
        dummy_file_path = make_dummy_file("test_template.sh", content=job_script_data_as_string)

        owner_email = job_script_data.owner_email
        requester_email = owner_email if is_owner else "another_" + owner_email

        inject_security_header(requester_email, *permissions)
        with open(dummy_file_path, mode="rb") as file:
            response = await client.put(
                f"jobbergate/job-scripts/{id}/upload/{file_type}",
                files={"upload_file": (dummy_file_path.name, file, "text/plain")},
            )

        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"

        # Check the response from the upload endpoint
        response_data = response.json()
        assert response_data is not None
        assert response_data["parent_id"] == id
        assert response_data["filename"] == dummy_file_path.name
        assert response_data["file_type"] == file_type

        # Check the database
        job_script_file = await synth_services.file.job_script.get(id, "test_template.sh")
        assert job_script_file is not None
        assert job_script_file.parent_id == id
        assert job_script_file.filename == dummy_file_path.name
        assert job_script_file.file_type == file_type
        assert job_script_file.file_key == f"job_script_files/{id}/{dummy_file_path.name}"

        # Check the file content on s3
        file_content = await synth_services.file.job_script.get_file_content(job_script_file)
        assert file_content.decode() == job_script_data_as_string

    @pytest.mark.parametrize(
        "is_owner, permissions, protocol",
        [
            (True, [Permissions.JOB_SCRIPTS_CREATE], "http"),
            (False, [Permissions.ADMIN], "http"),
            (False, [Permissions.JOB_SCRIPTS_CREATE, Permissions.MAINTAINER], "http"),
            (True, [Permissions.JOB_SCRIPTS_CREATE], "https"),
            (False, [Permissions.ADMIN], "https"),
            (False, [Permissions.JOB_SCRIPTS_CREATE, Permissions.MAINTAINER], "https"),
        ],
    )
    async def test_upsert_new_file_by_url__http_https(
        self,
        client,
        inject_security_header,
        job_script_data,
        synth_services,
        is_owner,
        permissions,
        protocol,
        respx_mock,
    ):
        dummy_content = "print('Hello World!')".encode()
        respx_mock.get(f"{protocol}://dummy-domain.com/dummy-file.py").mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                content=dummy_content,
            ),
        )

        id = job_script_data.id
        file_type = "ENTRYPOINT"

        owner_email = job_script_data.owner_email
        requester_email = owner_email if is_owner else "another_" + owner_email

        inject_security_header(requester_email, *permissions)
        response = await client.put(
            f"jobbergate/job-scripts/{id}/upload-by-url/{file_type}",
            params={"file_url": f"{protocol}://dummy-domain.com/dummy-file.py"},
        )

        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"

        # Check the response from the upload endpoint
        response_data = response.json()
        assert response_data is not None
        assert response_data["parent_id"] == id
        assert response_data["filename"] == "dummy-file.py"
        assert response_data["file_type"] == file_type

        # Check the database
        job_script_file = await synth_services.file.job_script.get(id, "dummy-file.py")
        assert job_script_file is not None
        assert job_script_file.parent_id == id
        assert job_script_file.filename == "dummy-file.py"
        assert job_script_file.file_type == file_type
        assert job_script_file.file_key == f"job_script_files/{id}/dummy-file.py"

        # Check the file content on s3
        file_content = await synth_services.file.job_script.get_file_content(job_script_file)
        assert file_content == dummy_content

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_SCRIPTS_CREATE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_SCRIPTS_CREATE, Permissions.MAINTAINER]),
        ],
    )
    async def test_upsert_new_file_by_url__s3(
        self,
        client,
        inject_security_header,
        job_script_data,
        synth_services,
        is_owner,
        permissions,
        respx_mock,
    ):
        dummy_content = "print('Hello World!')".encode()
        s3_url = "s3://dummy-bucket/dummy-file.py"
        https_url = "https://dummy-bucket.s3.amazonaws.com/dummy-file.py"
        respx_mock.get(https_url).mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                content=dummy_content,
            ),
        )

        id = job_script_data.id
        file_type = "ENTRYPOINT"

        owner_email = job_script_data.owner_email
        requester_email = owner_email if is_owner else "another_" + owner_email

        inject_security_header(requester_email, *permissions)
        response = await client.put(
            f"jobbergate/job-scripts/{id}/upload-by-url/{file_type}",
            params={"file_url": s3_url},
        )

        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"

        # Check the response from the upload endpoint
        response_data = response.json()
        assert response_data is not None
        assert response_data["parent_id"] == id
        assert response_data["filename"] == "dummy-file.py"
        assert response_data["file_type"] == file_type

        # Check the database
        job_script_file = await synth_services.file.job_script.get(id, "dummy-file.py")
        assert job_script_file is not None
        assert job_script_file.parent_id == id
        assert job_script_file.filename == "dummy-file.py"
        assert job_script_file.file_type == file_type
        assert job_script_file.file_key == f"job_script_files/{id}/dummy-file.py"

        # Check the file content on s3
        file_content = await synth_services.file.job_script.get_file_content(job_script_file)
        assert file_content == dummy_content

    @pytest.mark.parametrize("url_path", ["/blah/", "/", ""])
    async def test_upsert_new_file_by_url__fails_on_empty_filename_in_url(
        self,
        client,
        inject_security_header,
        job_script_data,
        url_path,
        respx_mock,
    ):
        dummy_content = "print('Hello World!')".encode()
        file_url = f"https://dummy-domain.com{url_path}"
        respx_mock.get(file_url).mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                content=dummy_content,
            ),
        )

        id = job_script_data.id
        file_type = "ENTRYPOINT"

        owner_email = job_script_data.owner_email
        requester_email = owner_email

        inject_security_header(requester_email, Permissions.JOB_SCRIPTS_CREATE)

        response = await client.put(
            f"jobbergate/job-scripts/{id}/upload-by-url/{file_type}",
            params={"file_url": file_url},
        )

        assert response.status_code == status.HTTP_400_BAD_REQUEST
        assert "Filename could not be extracted" in response.text

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_SCRIPTS_CREATE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_SCRIPTS_CREATE, Permissions.MAINTAINER]),
        ],
    )
    async def test_upsert_replace_content(
        self,
        client,
        inject_security_header,
        job_script_data,
        make_dummy_file,
        synth_services,
        is_owner,
        permissions,
    ):
        id = job_script_data.id
        file_type = "ENTRYPOINT"
        job_script_filename = "entrypoint.sh"

        await synth_services.file.job_script.upsert(
            parent_id=id,
            filename=job_script_filename,
            upload_content="original_content",
            file_type=file_type,
        )

        new_content = "new_content"
        dummy_file_path = make_dummy_file(job_script_filename, content=new_content)

        owner_email = job_script_data.owner_email
        requester_email = owner_email if is_owner else "another_" + owner_email
        inject_security_header(requester_email, *permissions)
        with open(dummy_file_path, mode="rb") as file:
            response = await client.put(
                f"jobbergate/job-scripts/{id}/upload/{file_type}",
                files={"upload_file": (dummy_file_path.name, file, "text/plain")},
            )

        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"

        job_script_file = await synth_services.file.job_script.get(id, job_script_filename)

        assert job_script_file is not None
        assert job_script_file.parent_id == id
        assert job_script_file.filename == job_script_filename
        assert job_script_file.file_type == file_type
        assert job_script_file.file_key == f"job_script_files/{id}/{dummy_file_path.name}"

        file_content = await synth_services.file.job_script.get_file_content(job_script_file)
        assert file_content.decode() == new_content

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_SCRIPTS_CREATE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_SCRIPTS_CREATE, Permissions.MAINTAINER]),
        ],
    )
    async def test_upsert_rename_file(
        self,
        client,
        inject_security_header,
        job_script_data,
        synth_services,
        is_owner,
        permissions,
    ):
        id = job_script_data.id
        file_type = "ENTRYPOINT"
        job_script_filename = "entrypoint.sh"

        await synth_services.file.job_script.upsert(
            parent_id=id,
            filename=job_script_filename,
            upload_content="original_content",
            file_type=file_type,
        )

        new_filename = "new_filename.sh"

        owner_email = job_script_data.owner_email
        requester_email = owner_email if is_owner else "another_" + owner_email
        inject_security_header(requester_email, *permissions)
        response = await client.put(
            f"jobbergate/job-scripts/{id}/upload/{file_type}",
            params={"filename": new_filename, "previous_filename": job_script_filename},
        )

        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"

        job_script_file = await synth_services.file.job_script.get(id, new_filename)

        assert job_script_file is not None
        assert job_script_file.parent_id == id
        assert job_script_file.filename == new_filename
        assert job_script_file.file_type == file_type
        assert job_script_file.file_key == f"job_script_files/{id}/{new_filename}"

        file_content = await synth_services.file.job_script.get_file_content(job_script_file)
        assert file_content.decode() == "original_content"

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_SCRIPTS_CREATE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_SCRIPTS_CREATE, Permissions.MAINTAINER]),
        ],
    )
    async def test_upsert_replace_content_and_rename(
        self,
        client,
        inject_security_header,
        job_script_data,
        make_dummy_file,
        synth_services,
        is_owner,
        permissions,
    ):
        id = job_script_data.id
        file_type = "ENTRYPOINT"
        job_script_filename = "entrypoint.sh"

        await synth_services.file.job_script.upsert(
            parent_id=id,
            filename=job_script_filename,
            upload_content="original_content",
            file_type=file_type,
        )

        new_content = "new_content"
        new_filename = "new_filename.sh"

        dummy_file_path = make_dummy_file(new_filename, content=new_content)
        owner_email = job_script_data.owner_email
        requester_email = owner_email if is_owner else "another_" + owner_email
        inject_security_header(requester_email, *permissions)
        with open(dummy_file_path, mode="rb") as file:
            response = await client.put(
                f"jobbergate/job-scripts/{id}/upload/{file_type}",
                files={"upload_file": (dummy_file_path.name, file, "text/plain")},
                params={
                    "filename": dummy_file_path.name,
                    "previous_filename": job_script_filename,
                },
            )

        assert response.status_code == status.HTTP_200_OK, f"Upsert failed: {response.text}"

        job_script_file = await synth_services.file.job_script.get(id, new_filename)

        assert job_script_file is not None
        assert job_script_file.parent_id == id
        assert job_script_file.filename == new_filename
        assert job_script_file.file_type == file_type
        assert job_script_file.file_key == f"job_script_files/{id}/{new_filename}"

        file_content = await synth_services.file.job_script.get_file_content(job_script_file)
        assert file_content.decode() == new_content

    async def test_create__fail_forbidden(
        self,
        client,
        tester_email,
        inject_security_header,
        job_script_data,
        job_script_data_as_string,
        make_dummy_file,
    ):
        id = job_script_data.id
        file_type = "ENTRYPOINT"
        dummy_file_path = make_dummy_file("test_template.sh", content=job_script_data_as_string)

        owner_email = tester_email
        requester_email = "another_" + owner_email

        inject_security_header(requester_email, Permissions.JOB_SCRIPTS_CREATE)
        with open(dummy_file_path, mode="rb") as file:
            response = await client.put(
                f"jobbergate/job-scripts/{id}/upload/{file_type}",
                files={"upload_file": (dummy_file_path.name, file, "text/plain")},
                params={"filename": dummy_file_path.name},
            )

        assert response.status_code == status.HTTP_403_FORBIDDEN

    @pytest.mark.parametrize("permission", (Permissions.ADMIN, Permissions.JOB_SCRIPTS_READ))
    async def test_get__success(
        self,
        permission,
        client,
        tester_email,
        inject_security_header,
        job_script_data,
        job_script_data_as_string,
        synth_services,
    ):
        id = job_script_data.id
        file_type = "ENTRYPOINT"
        job_script_filename = "entrypoint.sh"

        await synth_services.file.job_script.upsert(
            parent_id=id,
            filename=job_script_filename,
            upload_content=job_script_data_as_string,
            file_type=file_type,
        )

        inject_security_header(tester_email, permission)
        response = await client.get(f"jobbergate/job-scripts/{id}/upload/{job_script_filename}")

        assert response.status_code == status.HTTP_200_OK
        assert response.content.decode() == job_script_data_as_string

    async def test_get__large_file_success(
        self,
        client,
        tester_email,
        inject_security_header,
        job_script_data,
        synth_services,
    ):
        """
        Ensure that large files can be retrieved with no problem.
        This was created after strange issues were identified running on FastAPI 0.111.
        """
        large_string = "print(1)\n" * 5000

        id = job_script_data.id
        file_type = "ENTRYPOINT"
        job_script_filename = "entrypoint.sh"

        await synth_services.file.job_script.upsert(
            parent_id=id,
            filename=job_script_filename,
            upload_content=large_string,
            file_type=file_type,
        )

        inject_security_header(tester_email, Permissions.JOB_SCRIPTS_READ)
        response = await client.get(f"jobbergate/job-scripts/{id}/upload/{job_script_filename}")

        assert response.status_code == status.HTTP_200_OK, f"Get failed: {response.text}"
        assert response.content.decode() == large_string

    @pytest.mark.parametrize(
        "is_owner, permissions",
        [
            (True, [Permissions.JOB_SCRIPTS_DELETE]),
            (False, [Permissions.ADMIN]),
            (False, [Permissions.JOB_SCRIPTS_DELETE, Permissions.MAINTAINER]),
        ],
    )
    async def test_delete__success(
        self,
        client,
        inject_security_header,
        job_script_data,
        job_script_data_as_string,
        synth_services,
        synth_bucket,
        is_owner,
        permissions,
    ):
        parent_id = job_script_data.id
        file_type = "ENTRYPOINT"
        job_script_filename = "entrypoint.sh"

        upserted_instance = await synth_services.file.job_script.upsert(
            parent_id=parent_id,
            filename=job_script_filename,
            upload_content=job_script_data_as_string,
            file_type=file_type,
        )

        owner_email = job_script_data.owner_email
        requester_email = owner_email if is_owner else "another_" + owner_email

        inject_security_header(requester_email, *permissions)
        response = await client.delete(f"jobbergate/job-scripts/{parent_id}/upload/{job_script_filename}")
        assert response.status_code == status.HTTP_200_OK, f"Delete failed: {response.text}"

        s3_object = await synth_bucket.Object(upserted_instance.file_key)
        with pytest.raises(synth_bucket.meta.client.exceptions.NoSuchKey):
            await s3_object.get()

    async def test_delete__fail_forbidden(
        self,
        client,
        tester_email,
        inject_security_header,
        job_script_data,
        job_script_data_as_string,
        synth_services,
    ):
        parent_id = job_script_data.id
        file_type = "ENTRYPOINT"
        job_script_filename = "entrypoint.sh"

        await synth_services.file.job_script.upsert(
            parent_id=parent_id,
            filename=job_script_filename,
            upload_content=job_script_data_as_string,
            file_type=file_type,
        )

        owner_email = tester_email
        requester_email = "another_" + owner_email

        inject_security_header(requester_email, Permissions.JOB_SCRIPTS_DELETE)
        response = await client.delete(f"jobbergate/job-scripts/{parent_id}/upload/{job_script_filename}")
        assert response.status_code == status.HTTP_403_FORBIDDEN



"""create job metric table

Revision ID: 99c3877d0f10
Revises: 815022877cfe
Create Date: 2024-11-28 16:20:20.533111

"""

import time
from datetime import datetime, timezone
from textwrap import dedent

import sqlalchemy as sa
from sqlalchemy.types import TypeDecorator, DateTime

from alembic import op

# revision identifiers, used by Alembic.
revision = "99c3877d0f10"
down_revision = "815022877cfe"
branch_labels = None
depends_on = None


create_materialized_view_template = dedent(
    """
    CREATE MATERIALIZED VIEW {view_name}
    WITH (timescaledb.continuous) AS
    SELECT job_submission_id,
            node_host,
            time_bucket('{time_bucket}', time) AS bucket,
            AVG(cpu_frequency) AS cpu_frequency,
            SUM(cpu_time) AS cpu_time,
            AVG(cpu_utilization) AS cpu_utilization,
            AVG(gpu_memory) AS gpu_memory,
            AVG(gpu_utilization) AS gpu_utilization,
            SUM(page_faults) AS page_faults,
            AVG(memory_rss) AS memory_rss,
            AVG(memory_virtual) AS memory_virtual,
            SUM(disk_read) AS disk_read,
            SUM(disk_write) AS disk_write
    FROM job_submission_metrics
    GROUP BY job_submission_id, node_host, bucket
    WITH NO DATA
    """
)

add_continuous_aggregate_policy_template = dedent(
    """
    SELECT add_continuous_aggregate_policy('{mat_view_name}',
        start_offset => INTERVAL '{start_offset}',
        end_offset => INTERVAL '{end_offset}',
        schedule_interval => INTERVAL '{schedule_interval}',
        initial_start => '{initial_start}')
    """
)

remove_continuous_aggregate_policy_template = "SELECT remove_continuous_aggregate_policy('{mat_view_name}')"

drop_materialized_view_template = "DROP MATERIALIZED VIEW {view_name}"


class TimestampInt(TypeDecorator):
    impl = DateTime(timezone=True)

    def process_bind_param(self, value: int | None, dialect) -> datetime | None:
        if value is not None:
            return datetime.fromtimestamp(value, tz=timezone.utc)
        return value

    def process_result_value(self, value: datetime | None, dialect) -> int | None:
        if value is not None:
            return int(value.timestamp())
        return value


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(sa.text("CREATE EXTENSION IF NOT EXISTS timescaledb"))
    op.create_table(
        "job_submission_metrics",
        sa.Column(
            "time",
            TimestampInt(),
            nullable=False,
        ),
        sa.Column("job_submission_id", sa.Integer(), nullable=False),
        sa.Column("slurm_job_id", sa.Integer(), nullable=False),
        sa.Column("node_host", sa.String(), nullable=False),
        sa.Column("step", sa.Integer(), nullable=False),
        sa.Column("task", sa.Integer(), nullable=False),
        sa.Column("cpu_frequency", sa.Float(), nullable=False),
        sa.Column("cpu_time", sa.Float(), nullable=False),
        sa.Column("cpu_utilization", sa.Float(), nullable=False),
        sa.Column("gpu_memory", sa.BigInteger(), nullable=False),
        sa.Column("gpu_utilization", sa.Float(), nullable=False),
        sa.Column("page_faults", sa.BigInteger(), nullable=False),
        sa.Column("memory_rss", sa.BigInteger(), nullable=False),
        sa.Column("memory_virtual", sa.BigInteger(), nullable=False),
        sa.Column("disk_read", sa.BigInteger(), nullable=False),
        sa.Column("disk_write", sa.BigInteger(), nullable=False),
        sa.PrimaryKeyConstraint("time", "job_submission_id", "node_host", "step", "task"),
        sa.ForeignKeyConstraint(["job_submission_id"], ["job_submissions.id"], ondelete="CASCADE"),
    )
    op.create_index(
        "idx_job_submission_metrics_time",
        "job_submission_metrics",
        ["time"],
        unique=False,
    )
    op.create_index(
        "idx_node_host_step_task",
        "job_submission_metrics",
        ["node_host", "step", "task"],
        unique=False,
    )
    op.create_index(
        op.f("ix_job_submission_metrics_node_host"),
        "job_submission_metrics",
        ["node_host"],
        unique=False,
    )
    op.create_index(
        op.f("ix_job_submission_metrics_step"),
        "job_submission_metrics",
        ["step"],
        unique=False,
    )
    op.create_index(
        op.f("ix_job_submission_metrics_task"),
        "job_submission_metrics",
        ["task"],
        unique=False,
    )
    op.execute(sa.text("SELECT create_hypertable('job_submission_metrics', by_range('time'))"))
    op.execute(
        sa.text(
            create_materialized_view_template.format(
                view_name="metrics_nodes_mv_10_seconds_by_node",
                time_bucket="10 seconds",
            )
        )
    )
    op.execute(
        sa.text(
            add_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_10_seconds_by_node",
                start_offset="20 seconds",
                end_offset="0 seconds",
                schedule_interval="10 seconds",
                initial_start=datetime.now(timezone.utc).isoformat(),
            )
        )
    )
    op.execute(
        sa.text(
            create_materialized_view_template.format(
                view_name="metrics_nodes_mv_10_seconds_all_nodes",
                time_bucket="10 seconds",
            )
        )
    )
    op.execute(
        sa.text(
            add_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_10_seconds_all_nodes",
                start_offset="20 seconds",
                end_offset="0 seconds",
                schedule_interval="10 seconds",
                initial_start=datetime.now(timezone.utc).isoformat(),
            )
        )
    )
    op.execute(
        sa.text(
            create_materialized_view_template.format(
                view_name="metrics_nodes_mv_1_minute_by_node",
                time_bucket="1 minute",
            )
        )
    )
    op.execute(
        sa.text(
            add_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_1_minute_by_node",
                start_offset="2 minutes",
                end_offset="0 seconds",
                schedule_interval="1 minute",
                initial_start=datetime.now(timezone.utc).isoformat(),
            )
        )
    )
    op.execute(
        sa.text(
            create_materialized_view_template.format(
                view_name="metrics_nodes_mv_1_minute_all_nodes",
                time_bucket="1 minute",
            )
        )
    )
    op.execute(
        sa.text(
            add_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_1_minute_all_nodes",
                start_offset="2 minutes",
                end_offset="0 seconds",
                schedule_interval="1 minute",
                initial_start=datetime.now(timezone.utc).isoformat(),
            )
        )
    )
    op.execute(
        sa.text(
            create_materialized_view_template.format(
                view_name="metrics_nodes_mv_10_minutes_by_node",
                time_bucket="10 minutes",
            )
        )
    )
    op.execute(
        sa.text(
            add_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_10_minutes_by_node",
                start_offset="20 minutes",
                end_offset="0 seconds",
                schedule_interval="10 minutes",
                initial_start=datetime.now(timezone.utc).isoformat(),
            )
        )
    )
    op.execute(
        sa.text(
            create_materialized_view_template.format(
                view_name="metrics_nodes_mv_10_minutes_all_nodes",
                time_bucket="10 minutes",
            )
        )
    )
    op.execute(
        sa.text(
            add_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_10_minutes_all_nodes",
                start_offset="20 minutes",
                end_offset="0 seconds",
                schedule_interval="10 minutes",
                initial_start=datetime.now(timezone.utc).isoformat(),
            )
        )
    )
    op.execute(
        sa.text(
            create_materialized_view_template.format(
                view_name="metrics_nodes_mv_1_hour_by_node",
                time_bucket="1 hour",
            )
        )
    )
    op.execute(
        sa.text(
            add_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_1_hour_by_node",
                start_offset="2 hours",
                end_offset="0 seconds",
                schedule_interval="1 hour",
                initial_start=datetime.now(timezone.utc).isoformat(),
            )
        )
    )
    op.execute(
        sa.text(
            create_materialized_view_template.format(
                view_name="metrics_nodes_mv_1_hour_all_nodes",
                time_bucket="1 hour",
            )
        )
    )
    op.execute(
        sa.text(
            add_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_1_hour_all_nodes",
                start_offset="2 hours",
                end_offset="0 seconds",
                schedule_interval="1 hour",
                initial_start=datetime.now(timezone.utc).isoformat(),
            )
        )
    )
    op.execute(
        sa.text(
            create_materialized_view_template.format(
                view_name="metrics_nodes_mv_1_week_by_node",
                time_bucket="1 week",
            )
        )
    )
    op.execute(
        sa.text(
            add_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_1_week_by_node",
                start_offset="2 weeks",
                end_offset="0 seconds",
                schedule_interval="1 week",
                initial_start=datetime.now(timezone.utc).isoformat(),
            )
        )
    )
    op.execute(
        sa.text(
            create_materialized_view_template.format(
                view_name="metrics_nodes_mv_1_week_all_nodes",
                time_bucket="1 week",
            )
        )
    )
    op.execute(
        sa.text(
            add_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_1_week_all_nodes",
                start_offset="2 weeks",
                end_offset="0 seconds",
                schedule_interval="1 week",
                initial_start=datetime.now(timezone.utc).isoformat(),
            )
        )
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_job_submission_metrics_task"), table_name="job_submission_metrics")
    op.drop_index(op.f("ix_job_submission_metrics_step"), table_name="job_submission_metrics")
    op.drop_index(op.f("ix_job_submission_metrics_node_host"), table_name="job_submission_metrics")
    op.drop_index("idx_node_host_step_task", table_name="job_submission_metrics")
    op.drop_index("idx_job_submission_metrics_time", table_name="job_submission_metrics")
    op.execute(
        sa.text(
            remove_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_1_week_by_node"
            )
        )
    )
    op.execute(
        sa.text(
            remove_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_1_hour_by_node"
            )
        )
    )
    op.execute(
        sa.text(
            remove_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_10_minutes_by_node"
            )
        )
    )
    op.execute(
        sa.text(
            remove_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_1_minute_by_node"
            )
        )
    )
    op.execute(
        sa.text(
            remove_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_10_seconds_by_node"
            )
        )
    )
    op.execute(
        sa.text(
            remove_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_1_week_all_nodes"
            )
        )
    )
    op.execute(
        sa.text(
            remove_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_1_hour_all_nodes"
            )
        )
    )
    op.execute(
        sa.text(
            remove_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_10_minutes_all_nodes"
            )
        )
    )
    op.execute(
        sa.text(
            remove_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_1_minute_all_nodes"
            )
        )
    )
    op.execute(
        sa.text(
            remove_continuous_aggregate_policy_template.format(
                mat_view_name="metrics_nodes_mv_10_seconds_all_nodes"
            )
        )
    )
    op.execute(sa.text(drop_materialized_view_template.format(view_name="metrics_nodes_mv_1_week_by_node")))
    op.execute(sa.text(drop_materialized_view_template.format(view_name="metrics_nodes_mv_1_hour_by_node")))
    op.execute(
        sa.text(drop_materialized_view_template.format(view_name="metrics_nodes_mv_10_minutes_by_node"))
    )
    op.execute(sa.text(drop_materialized_view_template.format(view_name="metrics_nodes_mv_1_minute_by_node")))
    op.execute(
        sa.text(drop_materialized_view_template.format(view_name="metrics_nodes_mv_10_seconds_by_node"))
    )
    op.execute(sa.text(drop_materialized_view_template.format(view_name="metrics_nodes_mv_1_week_all_nodes")))
    op.execute(sa.text(drop_materialized_view_template.format(view_name="metrics_nodes_mv_1_hour_all_nodes")))
    op.execute(
        sa.text(drop_materialized_view_template.format(view_name="metrics_nodes_mv_10_minutes_all_nodes"))
    )
    op.execute(
        sa.text(drop_materialized_view_template.format(view_name="metrics_nodes_mv_1_minute_all_nodes"))
    )
    op.execute(
        sa.text(drop_materialized_view_template.format(view_name="metrics_nodes_mv_10_seconds_all_nodes"))
    )
    op.drop_table("job_submission_metrics")
    # ### end Alembic commands ###



"""Tests for the job submission service module."""

from itertools import product
from typing import Any, NamedTuple

import pendulum
import pytest
from sqlalchemy import inspect

from jobbergate_api.apps.constants import FileType


class TestIntegration:
    async def test_get_includes_all_files(
        self, fill_job_script_data, fill_job_submission_data, synth_services
    ):
        script_instance = await synth_services.crud.job_script.create(**fill_job_script_data())
        submission_instance = await synth_services.crud.job_submission.create(
            **fill_job_submission_data(), job_script_id=script_instance.id
        )
        script_file = await synth_services.file.job_script.upsert(
            script_instance.id,
            "test.txt",
            "test file content",
            file_type=FileType.ENTRYPOINT,
        )

        result = await synth_services.crud.job_submission.get(submission_instance.id, include_files=True)

        assert "job_script" not in inspect(result).unloaded
        assert "files" not in inspect(result.job_script).unloaded

        assert result.job_script.files == [script_file]

    async def test_get_includes_parent(self, fill_job_script_data, fill_job_submission_data, synth_services):
        script_instance = await synth_services.crud.job_script.create(**fill_job_script_data())
        submission_instance = await synth_services.crud.job_submission.create(
            **fill_job_submission_data(), job_script_id=script_instance.id
        )

        result = await synth_services.crud.job_submission.get(submission_instance.id, include_parent=True)

        assert "job_script" not in inspect(result).unloaded
        assert "files" in inspect(result.job_script).unloaded

        assert result.job_script == script_instance

    async def test_get_not_include_parent(
        self, fill_job_script_data, fill_job_submission_data, synth_services
    ):
        script_instance = await synth_services.crud.job_script.create(**fill_job_script_data())
        submission_instance = await synth_services.crud.job_submission.create(
            **fill_job_submission_data(), job_script_id=script_instance.id
        )

        result = await synth_services.crud.job_submission.get(submission_instance.id, include_parent=False)

        assert "job_script" in inspect(result).unloaded

    async def test_list_includes_all_files(
        self, fill_job_script_data, fill_job_submission_data, synth_services
    ):
        script_instance = await synth_services.crud.job_script.create(**fill_job_script_data())
        submission_instance = await synth_services.crud.job_submission.create(
            **fill_job_submission_data(), job_script_id=script_instance.id
        )
        script_file = await synth_services.file.job_script.upsert(
            script_instance.id,
            "test.txt",
            "test file content",
            file_type=FileType.ENTRYPOINT,
        )

        actual_result = await synth_services.crud.job_submission.list(include_files=True)

        assert actual_result == [submission_instance]
        assert actual_result[0].job_script.files == [script_file]

    async def test_update_includes_no_files(
        self, fill_job_script_data, fill_job_submission_data, synth_services
    ):
        script_instance = await synth_services.crud.job_script.create(**fill_job_script_data())
        submission_instance = await synth_services.crud.job_submission.create(
            **fill_job_submission_data(), job_script_id=script_instance.id
        )
        await synth_services.file.job_script.upsert(
            script_instance.id,
            "test.txt",
            "test file content",
            file_type=FileType.ENTRYPOINT,
        )

        result = await synth_services.crud.job_submission.update(submission_instance.id, name="new-name")

        actual_unloaded = inspect(result).unloaded
        expected_unloaded = {"job_script", "metrics", "progress_entries"}

        assert actual_unloaded == expected_unloaded


class TestEntryInfo(NamedTuple):
    """Named tuple to store the info on a test entry."""

    last_updated_delta: int
    is_archived: bool


def filter_test_entries(
    entries: dict[TestEntryInfo, dict[str, Any]],
    **kwargs: set[Any],
) -> set[int]:
    """
    This function returns a filter for test entries contained in a dictionary, based on specified target attributes.

    Given that entries are a dictionary of TestEntryInfo objects and their associated data,
    this function facilitates the retrieval of entry IDs that are contained in the set of values for each key in kwargs.
    """
    if not kwargs:
        return set()
    return {
        value["id"] for key, value in entries.items() if all(getattr(key, k) in v for k, v in kwargs.items())
    }


class TestAutoCleanUnusedJobSubmissions:
    """
    Test the clean_unused_entries method.
    """

    DAYS_TO_ARCHIVE = 1
    DAYS_TO_DELETE = 2

    @pytest.fixture
    def time_now(self):
        """
        Test fixture to freeze time for testing
        """
        time_now = pendulum.datetime(2025, 1, 1)
        with pendulum.travel_to(time_now, freeze=True):
            yield time_now

    @pytest.fixture
    async def dummy_data(
        self, fill_job_submission_data, time_now, synth_services
    ) -> dict[TestEntryInfo, dict[str, Any]]:
        """
        Create dummy test data covering a range of possible scenarios.

        The current time is pinned by pendulum, and the test data covers a range of days added to it.
        The test data is created with a range of possible values for last_updated_delta and is_archived.
        """
        result = {}

        LAST_UPDATED_DELTA_VALUES = (0, 1, 2)
        IS_ARCHIVED_VALUES = (True, False)

        for e in product(LAST_UPDATED_DELTA_VALUES, IS_ARCHIVED_VALUES):
            entry_info = TestEntryInfo(*e)
            data = fill_job_submission_data(is_archived=entry_info.is_archived)

            with pendulum.travel_to(time_now.add(days=entry_info.last_updated_delta), freeze=True):
                job_submission = await synth_services.crud.job_submission.create(**data)

            data["id"] = job_submission.id
            result[entry_info] = data

        return result

    async def test_auto_clean__unset(self, dummy_data, tweak_settings, time_now, synth_services):
        """
        Assert that nothing is deleted or archived when the thresholds are unset.
        """
        with (
            tweak_settings(
                AUTO_CLEAN_JOB_SUBMISSIONS_DAYS_TO_ARCHIVE=None,
                AUTO_CLEAN_JOB_SUBMISSIONS_DAYS_TO_DELETE=None,
            ),
            pendulum.travel_to(time_now.add(days=1_000), freeze=True),
        ):
            result = await synth_services.crud.job_submission.clean_unused_entries()

        assert result.archived == set()
        assert result.deleted == set()

        submissions_list = await synth_services.crud.job_submission.list()

        assert {s.id for s in submissions_list} == {s["id"] for s in dummy_data.values()}

    @pytest.mark.parametrize("time_delta", list(range(7)))
    async def test_auto_clean__set(
        self,
        time_delta,
        dummy_data,
        tweak_settings,
        time_now,
        synth_services,
    ):
        """
        Test the expected entries are archived or deleted when advancing the current time day by day.

        The test data is created with a range of possible values for last_updated_delta and is_archived.
        Considering the `DAYS_TO_ARCHIVE` and `DAYS_TO_DELETE` settings, we expect the following behavior:
        - On day zero, nothing should be archived or deleted.
          - Notice `set(range(0))` and `set(range(-1))` are empty sets,
            so no entries are returned by the filters.
        - On day one:
          - Entries not previously archived should be archived if last_updated_delta is 0.
          - No entries should be deleted.
        - On day two:
            - Entries not previously archived should be archived if last_updated_delta is 0 or 1.
            - Entries previously archived should be deleted if last_updated_delta is 0.
        - And so on for the following days.
        """
        with (
            tweak_settings(
                AUTO_CLEAN_JOB_SUBMISSIONS_DAYS_TO_ARCHIVE=self.DAYS_TO_ARCHIVE,
                AUTO_CLEAN_JOB_SUBMISSIONS_DAYS_TO_DELETE=self.DAYS_TO_DELETE,
            ),
            pendulum.travel_to(time_now.add(days=time_delta, minutes=1), freeze=True),
        ):
            result = await synth_services.crud.job_submission.clean_unused_entries()

        expected_archived_ids = filter_test_entries(
            dummy_data,
            is_archived={False},
            last_updated_delta=set(range(time_delta)),
        )
        assert result.archived == expected_archived_ids

        expected_deleted_ids = filter_test_entries(
            dummy_data,
            is_archived={True},
            last_updated_delta=set(range(time_delta - 1)),
        )
        assert result.deleted == expected_deleted_ids

        # Assert the deleted entries are not in the list of job submissions
        submissions_list = await synth_services.crud.job_submission.list()
        assert {s.id for s in submissions_list} == {s["id"] for s in dummy_data.values()} - result.deleted



"""Remap job submissions statuses

Revision ID: ec2d2948fb41
Revises: ba62c0fb9879
Create Date: 2024-01-30 10:52:38.671713

"""

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

from alembic import op

# revision identifiers, used by Alembic.
revision = "ec2d2948fb41"
down_revision = "ba62c0fb9879"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "job_submissions",
        sa.Column(
            "slurm_job_state",
            sa.Enum(
                "BOOT_FAIL",
                "CANCELLED",
                "COMPLETED",
                "CONFIGURING",
                "COMPLETING",
                "DEADLINE",
                "FAILED",
                "NODE_FAIL",
                "OUT_OF_MEMORY",
                "PENDING",
                "PREEMPTED",
                "RUNNING",
                "RESV_DEL_HOLD",
                "REQUEUE_FED",
                "REQUEUE_HOLD",
                "REQUEUED",
                "RESIZING",
                "REVOKED",
                "SIGNALING",
                "SPECIAL_EXIT",
                "STAGE_OUT",
                "STOPPED",
                "SUSPENDED",
                "TIMEOUT",
                "UNKNOWN",
                name="slurmjobstate",
                native_enum=False,
            ),
            nullable=True,
        ),
    )
    op.add_column("job_submissions", sa.Column("slurm_job_info", sa.String(), nullable=True))
    op.alter_column(
        "job_submissions",
        "status",
        existing_type=postgresql.ENUM(
            "CREATED",
            "SUBMITTED",
            "COMPLETED",
            "FAILED",
            "UNKNOWN",
            "REJECTED",
            "CANCELLED",
            name="jobsubmissionstatus",
        ),
        type_=sa.Enum(
            "CREATED",
            "SUBMITTED",
            "REJECTED",
            "DONE",
            "ABORTED",
            # -- Temporarily included until after remapping --
            "COMPLETED",
            "FAILED",
            "UNKNOWN",
            "CANCELLED",
            # ----
            name="jobsubmissionstatus",
            native_enum=False,
        ),
        existing_nullable=False,
    )

    # -- do the remapping --
    op.execute("UPDATE job_submissions SET status = 'DONE' WHERE status = 'COMPLETED'")
    op.execute("UPDATE job_submissions SET status = 'ABORTED' WHERE status = 'FAILED'")
    op.execute("UPDATE job_submissions SET status = 'ABORTED' WHERE status = 'UNKNOWN'")
    op.execute(
        """
        UPDATE job_submissions
        SET status = 'ABORTED',
            slurm_job_state = 'CANCELLED'
        WHERE status = 'CANCELLED'
    """
    )
    # ----

    op.alter_column(
        "job_submissions",
        "status",
        existing_type=postgresql.ENUM(
            "CREATED",
            "SUBMITTED",
            "REJECTED",
            "DONE",
            "ABORTED",
            "COMPLETED",
            "FAILED",
            "UNKNOWN",
            "CANCELLED",
            name="jobsubmissionstatus",
        ),
        type_=sa.Enum(
            "CREATED",
            "SUBMITTED",
            "REJECTED",
            "DONE",
            "ABORTED",
            name="jobsubmissionstatus",
            native_enum=False,
        ),
        existing_nullable=False,
    )


def downgrade():
    op.alter_column(
        "job_submissions",
        "status",
        existing_type=postgresql.ENUM(
            "CREATED",
            "SUBMITTED",
            "REJECTED",
            "DONE",
            "ABORTED",
            name="jobsubmissionstatus",
        ),
        type_=sa.Enum(
            "CREATED",
            "SUBMITTED",
            "COMPLETED",
            "FAILED",
            "UNKNOWN",
            "REJECTED",
            "CANCELLED",
            # -- Temporarily included until after remapping --
            "DONE",
            "ABORTED",
            # ----
            name="jobsubmissionstatus",
            native_enum=False,
        ),
        existing_nullable=False,
    )

    # -- do the remapping --
    op.execute(
        """
        UPDATE job_submissions
        SET status = 'CANCELLED'
        WHERE status = 'ABORTED'
        AND slurm_job_state = 'CANCELLED'
    """
    )
    op.execute("UPDATE job_submissions SET status = 'FAILED' WHERE status = 'ABORTED'")
    op.execute("UPDATE job_submissions SET status = 'COMPLETED' WHERE status = 'DONE'")
    # ----

    op.alter_column(
        "job_submissions",
        "status",
        existing_type=postgresql.ENUM(
            "CREATED",
            "SUBMITTED",
            "COMPLETED",
            "FAILED",
            "UNKNOWN",
            "REJECTED",
            "CANCELLED",
            "DONE",
            "ABORTED",
            name="jobsubmissionstatus",
        ),
        type_=sa.Enum(
            "CREATED",
            "SUBMITTED",
            "REJECTED",
            "DONE",
            "ABORTED",
            name="jobsubmissionstatus",
            native_enum=False,
        ),
        existing_nullable=False,
    )

    op.drop_column("job_submissions", "slurm_job_info")
    op.drop_column("job_submissions", "slurm_job_state")



"""Unit tests for retry utility functions."""

from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest

from jobbergate_api.retry_utils import async_retry, sync_retry


class TestAsyncRetry:
    """Tests for async_retry function."""

    @pytest.mark.asyncio
    async def test_success_on_first_attempt(self):
        """Test successful function execution on first attempt."""
        mock_func = AsyncMock(return_value="success")
        mock_func.__name__ = "test_func"
        result = await async_retry(mock_func)
        assert result == "success"
        assert mock_func.call_count == 1

    @pytest.mark.asyncio
    async def test_success_on_second_attempt(self):
        """Test successful function execution after one failure."""
        mock_func = AsyncMock(side_effect=[Exception("fail"), "success"])
        mock_func.__name__ = "test_func"
        result = await async_retry(mock_func, max_attempts=3)
        assert result == "success"
        assert mock_func.call_count == 2

    @pytest.mark.asyncio
    async def test_success_on_third_attempt(self):
        """Test successful function execution after two failures."""
        mock_func = AsyncMock(
            side_effect=[
                Exception("fail1"),
                Exception("fail2"),
                "success",
            ]
        )
        mock_func.__name__ = "test_func"
        result = await async_retry(mock_func, max_attempts=3)
        assert result == "success"
        assert mock_func.call_count == 3

    @pytest.mark.asyncio
    async def test_failure_after_max_attempts(self):
        """Test returnNone when all retries are exhausted."""
        mock_func = AsyncMock(side_effect=Exception("persistent failure"))
        mock_func.__name__ = "test_func"
        result = await async_retry(mock_func, max_attempts=3)
        assert result is None
        assert mock_func.call_count == 3

    @pytest.mark.asyncio
    async def test_exponential_backoff_timing(self):
        """Test that exponential backoff delays increase correctly."""
        mock_func = AsyncMock(side_effect=Exception("fail"))
        mock_func.__name__ = "test_func"

        with patch("jobbergate_api.retry_utils.asyncio.sleep", new_callable=AsyncMock) as mock_sleep:
            await async_retry(
                mock_func,
                max_attempts=4,
                initial_delay=1.0,
                backoff_factor=2.0,
            )

            # Should be called with 1.0, 2.0, 4.0 (three delays for 4 attempts)
            assert mock_sleep.call_count == 3
            calls = [call(delay) for delay in [1.0, 2.0, 4.0]]
            mock_sleep.assert_has_calls(calls)

    @pytest.mark.asyncio
    async def test_on_error_callback_invoked(self):
        """Test that on_error callback is invoked on each failure."""
        mock_func = AsyncMock(side_effect=Exception("fail"))
        mock_func.__name__ = "test_func"
        on_error_callback = MagicMock()

        await async_retry(
            mock_func,
            max_attempts=3,
            on_error=on_error_callback,
        )

        assert on_error_callback.call_count == 3
        # Verify attempt numbers are passed correctly (1, 2, 3)
        for call_args, expected_attempt in zip(on_error_callback.call_args_list, [1, 2, 3]):
            args, _ = call_args
            exc, attempt = args
            assert isinstance(exc, Exception)
            assert exc.args[0] == "fail"
            assert attempt == expected_attempt

    @pytest.mark.asyncio
    async def test_with_function_args_and_kwargs(self):
        """Test that args and kwargs are passed to the function."""
        mock_func = AsyncMock(return_value="result")
        mock_func.__name__ = "test_func"
        result = await async_retry(
            mock_func,
            "arg1",
            "arg2",
            kwarg1="value1",
            kwarg2="value2",
        )
        assert result == "result"
        mock_func.assert_called_once_with("arg1", "arg2", kwarg1="value1", kwarg2="value2")

    @pytest.mark.asyncio
    async def test_error_logging_on_exhaustion(self):
        """Test that error is logged when all retries are exhausted."""
        test_exception = Exception("test failure")
        mock_func = AsyncMock(side_effect=test_exception)
        mock_func.__name__ = "test_func"

        with patch("jobbergate_api.retry_utils.logger") as mock_logger:
            result = await async_retry(mock_func, max_attempts=2)
            assert result is None
            # Verify error was logged with exception details
            mock_logger.error.assert_called_once()
            error_message = mock_logger.error.call_args[0][0]
            assert "test failure" in error_message
            assert "2" in error_message  # max_attempts

    @pytest.mark.asyncio
    async def test_warning_logging_without_callback(self):
        """Test that warning is logged on each failure without callback."""
        mock_func = AsyncMock(side_effect=Exception("fail"))
        mock_func.__name__ = "test_func"

        with patch("jobbergate_api.retry_utils.logger") as mock_logger:
            await async_retry(mock_func, max_attempts=2)
            # Should log warning for each failure
            assert mock_logger.warning.call_count == 2

    @pytest.mark.asyncio
    async def test_no_warning_logging_with_callback(self):
        """Test that no warning is logged when callback is provided."""
        mock_func = AsyncMock(side_effect=Exception("fail"))
        mock_func.__name__ = "test_func"
        on_error_callback = MagicMock()

        with patch("jobbergate_api.retry_utils.logger") as mock_logger:
            await async_retry(
                mock_func,
                max_attempts=2,
                on_error=on_error_callback,
            )
            # Should not log warning when callback is provided
            mock_logger.warning.assert_not_called()


class TestSyncRetry:
    """Tests for sync_retry function."""

    def test_success_on_first_attempt(self):
        """Test successful function execution on first attempt."""
        mock_func = MagicMock(return_value="success")
        mock_func.__name__ = "test_func"
        result = sync_retry(mock_func)
        assert result == "success"
        assert mock_func.call_count == 1

    def test_success_on_second_attempt(self):
        """Test successful function execution after one failure."""
        mock_func = MagicMock(side_effect=[Exception("fail"), "success"])
        mock_func.__name__ = "test_func"
        result = sync_retry(mock_func, max_attempts=3)
        assert result == "success"
        assert mock_func.call_count == 2

    def test_success_on_third_attempt(self):
        """Test successful function execution after two failures."""
        mock_func = MagicMock(
            side_effect=[
                Exception("fail1"),
                Exception("fail2"),
                "success",
            ]
        )
        mock_func.__name__ = "test_func"
        result = sync_retry(mock_func, max_attempts=3)
        assert result == "success"
        assert mock_func.call_count == 3

    def test_failure_after_max_attempts(self):
        """Test return None when all retries are exhausted."""
        mock_func = MagicMock(side_effect=Exception("persistent failure"))
        mock_func.__name__ = "test_func"

        with patch("time.sleep"):
            result = sync_retry(mock_func, max_attempts=3)

        assert result is None
        assert mock_func.call_count == 3

    def test_exponential_backoff_timing(self):
        """Test that exponential backoff delays increase correctly."""
        mock_func = MagicMock(side_effect=Exception("fail"))
        mock_func.__name__ = "test_func"

        with patch("time.sleep") as mock_sleep:
            sync_retry(
                mock_func,
                max_attempts=4,
                initial_delay=1.0,
                backoff_factor=2.0,
            )

            # Should be called with 1.0, 2.0, 4.0 (three delays for 4 attempts)
            assert mock_sleep.call_count == 3
            calls = [call(delay) for delay in [1.0, 2.0, 4.0]]
            mock_sleep.assert_has_calls(calls)

    def test_on_error_callback_invoked(self):
        """Test that on_error callback is invoked on each failure."""
        mock_func = MagicMock(side_effect=Exception("fail"))
        mock_func.__name__ = "test_func"
        on_error_callback = MagicMock()

        with patch("time.sleep"):
            sync_retry(
                mock_func,
                max_attempts=3,
                on_error=on_error_callback,
            )

        assert on_error_callback.call_count == 3
        # Verify attempt numbers are passed correctly (1, 2, 3)
        for call_args, expected_attempt in zip(on_error_callback.call_args_list, [1, 2, 3]):
            args, _ = call_args
            exc, attempt = args
            assert isinstance(exc, Exception)
            assert exc.args[0] == "fail"
            assert attempt == expected_attempt

    def test_with_function_args_and_kwargs(self):
        """Test that args and kwargs are passed to the function."""
        mock_func = MagicMock(return_value="result")
        mock_func.__name__ = "test_func"
        result = sync_retry(
            mock_func,
            "arg1",
            "arg2",
            kwarg1="value1",
            kwarg2="value2",
        )
        assert result == "result"
        mock_func.assert_called_once_with("arg1", "arg2", kwarg1="value1", kwarg2="value2")

    def test_error_logging_on_exhaustion(self):
        """Test that error is logged when all retries are exhausted."""
        test_exception = Exception("test failure")
        mock_func = MagicMock(side_effect=test_exception)
        mock_func.__name__ = "test_func"

        with patch("jobbergate_api.retry_utils.logger") as mock_logger:
            with patch("time.sleep"):
                result = sync_retry(mock_func, max_attempts=2)

            assert result is None
            # Verify error was logged with exception details
            mock_logger.error.assert_called_once()
            error_message = mock_logger.error.call_args[0][0]
            assert "test failure" in error_message
            assert "2" in error_message  # max_attempts

    def test_warning_logging_without_callback(self):
        """Test that warning is logged on each failure without callback."""
        mock_func = MagicMock(side_effect=Exception("fail"))
        mock_func.__name__ = "test_func"

        with patch("jobbergate_api.retry_utils.logger") as mock_logger:
            with patch("time.sleep"):
                sync_retry(mock_func, max_attempts=2)

            # Should log warning for each failure
            assert mock_logger.warning.call_count == 2

    def test_no_warning_logging_with_callback(self):
        """Test that no warning is logged when callback is provided."""
        mock_func = MagicMock(side_effect=Exception("fail"))
        mock_func.__name__ = "test_func"
        on_error_callback = MagicMock()

        with patch("jobbergate_api.retry_utils.logger") as mock_logger:
            with patch("time.sleep"):
                sync_retry(
                    mock_func,
                    max_attempts=2,
                    on_error=on_error_callback,
                )

            # Should not log warning when callback is provided
            mock_logger.warning.assert_not_called()



"""
Email notification system for Jobbergate.
"""

from dataclasses import dataclass
from typing import List, Optional, Union

from buzz import Buzz
from loguru import logger
from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Mail

from jobbergate_api.config import settings
from jobbergate_api.retry_utils import sync_retry


class EmailNotificationError(Buzz):
    """
    Custom error to be raised for problems at the email notification system.
    """

    pass


@dataclass
class EmailManager:
    """
    Email manager.
    """

    email_client: SendGridAPIClient
    from_email: Optional[str] = ""

    def send_email(
        self,
        to_emails: Union[str, List[str]],
        subject: str,
        skip_on_failure: bool = False,
        enable_retry: bool = False,
        **kwargs,
    ) -> bool:
        """
        Send an email using this manager.

        Args:
            to_emails: Email or list of emails to send to
            subject: Email subject
            skip_on_failure: If True, log errors instead of raising exceptions
            enable_retry: If True, attempt delivery with retry logic (3 attempts, exponential backoff)
            **kwargs: Additional arguments for Mail construction

        Returns:
            True if email was sent successfully, False if skipped on failure
        """

        def _send() -> None:
            with EmailNotificationError.handle_errors(
                f"Error while sending email from_email={self.from_email}, {to_emails=}",
                do_except=lambda params: logger.warning(params.final_message),
                re_raise=not skip_on_failure,
            ):
                message = self._build_message(to_emails, subject, **kwargs)
                self.email_client.send(message)

        if enable_retry:

            def on_retry_error(exc: Exception, attempt: int) -> None:
                logger.warning(f"Failed to send email (attempt {attempt}): {exc}")

            result = sync_retry(
                _send,
                max_attempts=3,
                initial_delay=1.0,
                backoff_factor=2.0,
                on_error=on_retry_error,
            )
            if result is None:
                logger.error("Failed to send email after 3 retry attempts")
                return False
        else:
            try:
                _send()
            except EmailNotificationError:
                if skip_on_failure:
                    return False
                raise
        return True

    def _build_message(self, to_emails: Union[str, List[str]], subject: str, **kwargs) -> Mail:
        """
        Build an email message.
        """
        EmailNotificationError.require_condition(self.from_email, "The value from_email is empty.")

        with EmailNotificationError.handle_errors(
            "Error while creating the message",
            do_except=lambda params: logger.error(params.final_message),
        ):
            message = Mail(
                from_email=self.from_email,
                to_emails=to_emails,
                subject=subject,
                **kwargs,
            )
        return message


def notify_submission_rejected(
    job_submission_id: Union[str, int], report_message: str, to_emails: Union[str, List[str]]
) -> None:
    """
    Notify an email or a list of emails about a job submission that has been rejected.
    """
    subject = f"Job Submission Rejected (id={job_submission_id})"

    logger.debug(f"Notifying {to_emails=} that {job_submission_id=} was rejected: {report_message=}")
    email_manager.send_email(to_emails, subject, skip_on_failure=True, plain_text_content=report_message)


email_client = SendGridAPIClient(settings.SENDGRID_API_KEY)
email_manager = EmailManager(
    email_client,
    settings.SENDGRID_FROM_EMAIL,
)



"""Add clone capability

Revision ID: ba62c0fb9879
Revises: 44d41f873c8d
Create Date: 2024-01-15 17:11:10.626366

"""

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

from alembic import op

# revision identifiers, used by Alembic.
revision = "ba62c0fb9879"
down_revision = "44d41f873c8d"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("job_script_templates", sa.Column("cloned_from_id", sa.Integer(), nullable=True))
    op.create_foreign_key(
        "clone",
        "job_script_templates",
        "job_script_templates",
        ["cloned_from_id"],
        ["id"],
        ondelete="SET NULL",
    )
    op.add_column("job_scripts", sa.Column("cloned_from_id", sa.Integer(), nullable=True))
    op.create_foreign_key(
        "clone", "job_scripts", "job_scripts", ["cloned_from_id"], ["id"], ondelete="SET NULL"
    )
    op.add_column("job_submissions", sa.Column("cloned_from_id", sa.Integer(), nullable=True))
    op.create_foreign_key(
        "clone", "job_submissions", "job_submissions", ["cloned_from_id"], ["id"], ondelete="SET NULL"
    )


def downgrade():
    op.drop_constraint("clone", "job_submissions", type_="foreignkey")
    op.drop_column("job_submissions", "cloned_from_id")
    op.drop_constraint("clone", "job_scripts", type_="foreignkey")
    op.drop_column("job_scripts", "cloned_from_id")
    op.drop_constraint("clone", "job_script_templates", type_="foreignkey")
    op.drop_column("job_script_templates", "cloned_from_id")
